#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2012-2018, SSI. All rights reserved.

import indigo
import locale
import sys
import os
from os import listdir
import subprocess
import time
import socket
import urllib
import urllib2
import shutil
import simplejson as json
import logging
from contextlib import closing
import BaseHTTPServer
from SimpleHTTPServer import SimpleHTTPRequestHandler
from threading import Thread
from xml.etree import ElementTree as ET
import re

from AppKit import NSSpeechSynthesizer
from AppKit import NSURL

sys.path.insert(1, './lib')
import requests
from twisted.internet import reactor
from twisted.internet.protocol import DatagramProtocol
from twisted.application.internet import MulticastServer
from gtts import gTTS
import pyvona
import boto3
#from elementtree import ElementTree as ET
from mutagen.mp3 import MP3
from mutagen.aiff import AIFF
from pandora import Pandora

import soco
from soco import config
from soco import SoCo
from soco.music_services import MusicService

import language_codes

SONOS_ZonePlayer = 0
SONOS_CONNECT = 1
SONOS_CONNECTAMP = 2
SONOS_PLAY1 = 3
SONOS_PLAY3 = 4
SONOS_PLAY5 = 5
SONOS_PLAYBAR = 6
SONOS_SUB = 7

ZP_LIST = []

Sonos_Favorites = []
Sonos_Playlists = []
Sonos_RT_Fav_Stations = []
Sonos_LineIn = []
Sonos_Pandora = []
Sonos_SiriusXM = []

SavedState = []

Sound_Files = []

ContainerUpdateID_SQ = 0

# used to bypass ZonePlayer updates for certain actions like Group Announcements
actionBusy = 0

UPNP_ERRORS = {
	'400': 'Bad Request',
	'401': 'Invalid Action',
	'402': 'Invalid Args',
	'404': 'Invalid Var',
	'412': 'Precondition Failed',
	'501': 'Action Failed',
	'600': 'Argument Value Invalid',
	'601': 'Argument Value Out of Range',
	'602': 'Optional Action Not Implemented',
	'603': 'Out of Memory',
	'604': 'Human Invtervention Required',
	'605': 'String Argument Too Long',
	'606': 'Action Not Authorized',
	'607': 'Signature Failure',
	'608': 'Signature Missing',
	'609': 'Not Encrypted',
	'610': 'Invalid Sequence',
	'611': 'Invalid Control URL',
	'612': 'No Such Session',
	'701': 'No Such Object',
	'702': 'Invalid CurrentTagValue',
	'703': 'Invalid NewTagValue',
	'704': 'RequiredTag',
	'705': 'Read Only Tag',
	'706': 'Parameter Mismatch',
	'708': 'Unsupported or Invalid Search Criteria',
	'709': 'Unsupported or Invalid Sort Criteria',
	'710': 'No Such Container',
	'711': 'Restricted Object',
	'712': 'Bad Metadata',
	'713': 'Restricted Parent Object',
	'714': 'No Such Source Resource',
	'715': 'Resource Access Denied',
	'716': 'Transfer Busy',
	'717': 'No Such File Transfer',
	'718': 'No Such Destination Resource',
	'719': 'Destination Resource Access Denied',
	'720': 'Cannot Process the Request',
	'800': 'Unable to Play the Selected Item',
	'804': 'Invalid Queue Request'
	}

uri_tv = "x-sonos-htastream:"
uri_music = "x-rincon-queue:"
uri_radio = "x-sonosapi-stream:"
uri_sonos_radio = "x-sonosapi-radio:"
uri_sonos_http = "x-sonosapi-http:"
uri_pandora = "pndrradio:"
uri_siriusxm = "x-sonosapi-hls:"

uri_jffs = "file:"
uri_file = "x-file-cifs"
uri_group = "x-rincon:"
uri_playlist = "x-rincon-playlist:"
uri_mp3radio = "x-rincon-mp3radio:"
uri_container = "x-rincon-cpcontainer"

ZoneGroupStates = set(['ZP_ALBUM','ZP_ARTIST','ZP_CREATOR','ZP_TRACK','ZP_NALBUM','ZP_NART','ZP_NARTIST','ZP_NCREATOR','ZP_NTRACK',\
			'ZP_CurrentTrack','ZP_CurrentTrackURI','ZP_DURATION','ZP_INFO','ZP_STATION','ZP_STATE'])

IVONAlanguages = {'en-US':'English, American',
	'en-AU':'English, Australian',
	'en-GB':'English, British',
	'en-IN':'English, Indian',
	'en-GB-WLS':'English, Welsh',
	'cy-GB':'Welsh',
	'da-DK':'Danish',
	'nl-NL':'Dutch',
	'fr-FR':'French',
	'fr-CA':'French, Canadian',
	'de-DE':'German',
	'is-IS':'Icelandic',
	'it-IT':'Italian',
	'pl-PL':'Polish',
	'pt-PT':'Portuguese',
	'pt-BR':'Portuguese, Brazilian',
	'ro-RO':'Romanian',
	'ru-RU':'Russian',
	'es-ES':'Spanish, Castilian',
	'es-US':'Spanish, American',
	'sv-SE':'Swedish',
	'tr-TR':'Turkish',
	'nb-NO':'Norweigian'
	}
## IVONAlanguages['en-US'] = 'English, American'

IVONAVoices = []
PollyVoices = []
NSVoices = []

class PA():
	def __init__(self, deviceId=None, props=None):
		self.deviceId = deviceId
		self.props = props

class SSDPListener(DatagramProtocol):
	def __init__(self, cls):
		self.cls = cls

	def startProtocol(self):
		self.transport.setTTL(5)
		self.transport.joinGroup("239.255.255.250")
		indigo.server.log("SSDP Listener Started...")

	def datagramReceived(self, data, (address, port)):
		self.SSDPProcess(address, data)

	def SSDPProcess(self, address, data):
		message = {}
		message['address'] = address

		if ("Sonos" in data) and ("NOTIFY" in data):
			for line in data.split('\r\n'):
				if (len(line) > 0) and ("NOTIFY" not in line):
					(item, value)  = line.split(': ')
					message[item] = value
			
			for deviceId in self.cls.deviceList:
				dev = indigo.devices[int(deviceId)]
				if dev.address == message['address']:
					if (dev.states["ZP_LocalUID"] == message['NT'][5:]) and (dev.states["ZP_LocalUID"] == message['USN'][5:]):
						if message['NTS'] == 'ssdp:byebye':
							self.cls.plugin.errorLog("[%s] Network Error, Disconnect from ZonePlayer: %s" % (time.asctime(), dev.name))
							dev.setErrorStateOnServer(u"error")
						elif message['NTS'] == 'ssdp:alive':
							self.cls.updateStateOnServer(dev, "alive", time.asctime())
							self.cls.plugin.debugLog("[%s] Received ALIVE message from ZonePlayer: %s" % (time.asctime(), dev.name))
							if int(message['X-RINCON-BOOTSEQ']) != int(dev.states["bootseq"]):
								self.cls.socoResubscribe(dev, message['X-RINCON-BOOTSEQ'])
					break

class Sonos(object):

	######################################################################################
	# class init & del
	def __init__(self, plugin):
		self.plugin = plugin
		self.EventProcessor = None
		self.EventIP = None
		self.HTTPStreamingIP = None
		self.HTTPStreamingPort = None
		self.HTTPStreamerOn = None
		self.httpd = None
		self.SoundFilePath = None
		self.rootZPIP = None
		self.Pandora = None
		self.PandoraEmailAddress = None
		self.PandoraPassword = None
		self.PandoraNickName = None
		self.Pandora2 = None
		self.PandoraEmailAddress2 = None
		self.PandoraPassword2 = None
		self.PandoraNickname2 = None
		self.SiriusXM = None
		self.SiriusXMID = None
		self.SiriusXMPassword = None
		self.IVONA = None
		self.IVONAaccessKey = None
		self.IVONAsecretKey = None
		self.Polly = None
		self.PollyaccessKey = None
		self.PollysecretKey = None
		self.MSTranslate = None
		self.MSTranslateClientID = None
		self.MSTranslateClientSecret = None
		self.ttsORfile = None
		self.deviceList = []
		self.ZonePlayers = []
		self.ZPTypes = []

		self.SonArray = [{}]
		self.HTTPServer = None
		self.control_point = None

		self.soco_sub = {}
		self.event_threads = {}

		self.MSTranslateVoices = {}
		self.myLocale = None

		self.zonePlayerState = {}
		self.SonosDeviceID = None

	def __del__(self):
		pass

	def startup(self):
		global NSVoices

		self.plugin.debugLog("Sonos Plugin Startup...")

		logging.getLogger("requests").setLevel(logging.WARNING)
		#logging.getLogger("soco").setLevel(logging.INFO)

		if os.path.exists ('/Library/Application Support/Perceptive Automation/images/Sonos') == 0:
			os.makedirs('/Library/Application Support/Perceptive Automation/images/Sonos')

		self.closedPrefsConfigUi(None, None)

		try:
			NSVoices = NSSpeechSynthesizer.availableVoices()
			indigo.server.log(u"Loaded Apple Voices.. [%s]" % len(NSVoices))
		except:
			self.plugin.errorLog("[%s] Cannot load Apple Voices." % time.asctime())

		self.plugin.updater.checkVersionPoll()

		reactor.listenMulticast(1900, SSDPListener(self), listenMultiple=True)
		soco.config.EVENT_LISTENER_IP = self.EventIP

	def shutdown(self):
		self.HTTPStreamerOn = False
		if self.EventProcessor == "SoCo":
			indigo.server.log("SoCo Reactor Landing...")
			if reactor.running:
				try:
					reactor.stop()
				except:
					pass


	def deviceStartComm(self, device):
		self.plugin.debugLog("ZonePlayer: %s, Enabled: %s" % (device.name, device.enabled))
		if device.id not in self.deviceList:
			self.deviceList.append(device.id)
			device.stateListOrDisplayStateIdChanged()
			self.initZones(device)

	def deviceStopComm(self, device):
		self.plugin.debugLog("Stopping device: %s " % device.name)
		if device.id in self.deviceList:
			self.deviceList.remove(device.id)
			if self.EventProcessor == "SoCo":
				self.socoUnsubscribe (device)

	def SoCoReactor(self):
		indigo.server.log("SoCo Reactor Ignition...")
		reactor.run(installSignalHandlers=0)

	def HTTPStreamer(self):
		try:
			HandlerClass = SimpleHTTPRequestHandler
			ServerClass = BaseHTTPServer.HTTPServer
			Protocol = "HTTP/1.0"

			if self.HTTPStreamingIP == "auto":
				self.HTTPServer = socket.gethostbyname(socket.gethostname())
				if self.HTTPServer == None:
					d = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
					d.connect(("indigodomo.com",80))
					self.HTTPServer = d.getsockname()[0]
					d.close
			else:
				self.HTTPServer = self.HTTPStreamingIP

			server_address = ('0.0.0.0', int(self.HTTPStreamingPort))
			HandlerClass.protocol_version = Protocol
			self.httpd = ServerClass(server_address, HandlerClass)

			sa = self.httpd.socket.getsockname()
			indigo.server.log ("Serving HTTP Streamer on %s [%s], port %s" % (self.HTTPServer, sa[0], sa[1]))
			self.HTTPStreamerOn = True
			while self.HTTPStreamerOn:
				self.httpd.handle_request()
		except Exception, e:
			self.plugin.errorLog ("Cannot start HTTP Streamer on %s: %s" % (self.HTTPServer, e))


	######################################################################################
	# Concurrent Thread Start
	def runConcurrentThread(self):

		self.plugin.debugLog(u"Running Concurrent Thread")

		if self.EventProcessor == "SoCo":
			x = Thread(target=self.SoCoReactor)
			x.setDaemon(True)
			x.start()

		updateTime = time.time()
		subTestTime = time.time()

		while self.plugin.StopThread == False:

			if time.time() - updateTime >= 60:
				self.plugin.updater.checkVersionPoll()
				updateTime = time.time()
				self.checkEvents()

			if time.time() - subTestTime >= 15:
				subTestTime = time.time()
				self.socoSubTest()

			self.plugin.sleep(0.1)

	def restartPlugin(self):
		indigo.server.log("Restarting Sonos Plugin...")
		self.plugin.stopConcurrentThread()
		SonosPlug = indigo.server.getPlugin("com.ssi.indigoplugin.Sonos")
		SonosPlug.restart (waitUntilDone=True)

	def stopConcurrentThread(self):
		self.plugin.debugLog(u"Stopping Concurrent Thread")
		self.plugin.StopThread = True

	def checkEvents(self):
		timeout = (15 * 60) + 10
		for dev in indigo.devices.iter("self.ZonePlayer"):
			if dev.enabled:
				try:
					if time.time() - time.mktime(time.strptime(dev.states["alive"], "%a %b %d %H:%M:%S %Y")) > timeout:
						self.plugin.errorLog("[%s] ZonePlayer: %s has fallen off the network." % (time.asctime(), dev.name))
						dev.setErrorStateOnServer(u"error")
						self.socoTimeout (dev)
				except:
						dev.setErrorStateOnServer(u"error")
						self.socoTimeout (dev)
					

	######################################################################################
	# Check for messages
	def initZones(self, dev):
		zoneIP = dev.pluginProps["address"]
		self.plugin.debugLog(u"Resetting States for zone: %s" % zoneIP)
		self.updateStateOnServer (dev, "ZP_ALBUM", "")
		self.updateStateOnServer (dev, "ZP_ART", "")
		self.updateStateOnServer (dev, "ZP_ARTIST", "")
		self.updateStateOnServer (dev, "ZP_CREATOR", "")
		self.updateStateOnServer (dev, "ZP_CurrentURI", "")
		self.updateStateOnServer (dev, "ZP_DURATION", "")
		self.updateStateOnServer (dev, "ZP_INFO", "")
		self.updateStateOnServer (dev, "ZP_MUTE", "")
		self.updateStateOnServer (dev, "ZP_STATE", "")
		self.updateStateOnServer (dev, "ZP_STATION", "")
		self.updateStateOnServer (dev, "ZP_TRACK", "")
		self.updateStateOnServer (dev, "ZP_VOLUME", "")
		self.updateStateOnServer (dev, "ZP_VOLUME_FIXED", "")
		self.updateStateOnServer (dev, "ZP_BASS", "")
		self.updateStateOnServer (dev, "ZP_TREBLE", "")
		self.updateStateOnServer (dev, "ZP_ZoneName", "")
		self.updateStateOnServer (dev, "ZP_LocalUID", "")
		self.updateStateOnServer (dev, "ZP_AIName", "")
		self.updateStateOnServer (dev, "ZP_AIPath", "")
		self.updateStateOnServer (dev, "ZP_NALBUM", "")
		self.updateStateOnServer (dev, "ZP_NART", "")
		self.updateStateOnServer (dev, "ZP_NARTIST", "")
		self.updateStateOnServer (dev, "ZP_NCREATOR", "")		
		self.updateStateOnServer (dev, "ZP_NTRACK", "")
		self.updateStateOnServer (dev, "Q_Crossfade", "off")
		self.updateStateOnServer (dev, "Q_Repeat", "off")
		self.updateStateOnServer (dev, "Q_RepeatOne", "off")
		self.updateStateOnServer (dev, "Q_Shuffle", "off")
		self.updateStateOnServer (dev, "Q_Number", "0")
		self.updateStateOnServer (dev, "Q_ObjectID", "")
		self.updateStateOnServer (dev, "GROUP_Coordinator", "")
		self.updateStateOnServer (dev, "GROUP_Name", "")
		self.updateStateOnServer (dev, "ZP_CurrentTrack", "")
		self.updateStateOnServer (dev, "ZP_CurrentTrackURI", "")
		self.updateStateOnServer (dev, "ZoneGroupID", "")
		self.updateStateOnServer (dev, "ZoneGroupName", "")
		self.updateStateOnServer (dev, "ZonePlayerUUIDsInGroup", "")
		self.updateStateOnServer (dev, "alive", "")
		self.updateStateOnServer (dev, "bootseq", "")

		url = u"http://" + zoneIP + ":1400/status/zp"
		response = requests.get(url)
		if response.ok:
			root = ET.fromstring(response.content)
			ZoneName = root.findtext('.//ZoneName')
			LocalUID = root.findtext('.//LocalUID')
			SerialNumber = root.findtext('.//SerialNumber')
		else:
			self.plugin.errorLog("Error getting ZonePlayer data: %s" % url)
			self.plugin.errorLog("  Offending ZonePlayer: %s" % dev.name)
			self.plugin.errorLog("  ZonePlayer may be physically turned off or in a bad state.")
			self.plugin.errorLog("  Please disable communications or remove from Indigo.")
			dev.setErrorStateOnServer(u"error")
			return

		#self.updateStateOnServer (dev, "ZP_ZoneName", ZPInfo.findtext('ZoneName').decode('utf-8'))
		# Allow for special characters in ZoneName
		self.updateStateOnServer (dev, "ZP_ZoneName", ZoneName)
		self.updateStateOnServer (dev, "ZP_LocalUID", LocalUID)
		self.updateStateOnServer (dev, "SerialNumber", SerialNumber)

		self.getModelName (dev)

		self.updateZoneGroupStates (dev)
		self.updateZoneTopology (dev)

		indigo.server.log ("Adding ZonePlayer: %s, %s, %s" % (zoneIP, LocalUID, dev.name))
		self.ZonePlayers.append (LocalUID)
		self.ZPTypes.append ([LocalUID, dev.pluginProps["model"]])

		self.zonePlayerState[dev.id] = {'zonePlayerAlive':True}
		self.updateStateOnServer (dev, "alive", time.asctime())

		if self.EventProcessor == "SoCo":
			self.socoSubscribe (dev)

	def socoTimeout(self, dev):
		self.plugin.errorLog("[%s] Resubscription due to timeout for ZonePlayer: %s" % (time.asctime(), dev.name))
		self.socoUnsubscribe (dev)
		self.socoSubscribe (dev)

	def socoSubTest(self):
		#indigo.server.log("[%s] socoSubTest" % time.asctime())
		for dev in indigo.devices.iter("self.ZonePlayer"):
			if dev.enabled:
				subStatus = True
				if self.soco_sub[dev.id]['avTransport'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: avTransport subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if self.soco_sub[dev.id]['renderingControl'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: renderingControl subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if self.soco_sub[dev.id]['zoneGroupTopology'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: zoneGroupTopology subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if self.soco_sub[dev.id]['queue'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: queue subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if self.soco_sub[dev.id]['groupRenderingControl'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: groupRenderingControl subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if self.soco_sub[dev.id]['contentDirectory'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: contentDirectory subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if self.soco_sub[dev.id]['groupManagement'].is_subscribed != True:
					self.plugin.errorLog("[%s] [%s]: groupManagement subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
					subStatus = False
				if int(dev.pluginProps["model"]) in [SONOS_CONNECT, SONOS_CONNECTAMP, SONOS_PLAY5]:
					if self.soco_sub[dev.id]['audioIn'].is_subscribed != True:
						self.plugin.errorLog("[%s] [%s]: audioIn subscription failed.  Resubscribing..." % (time.asctime(), dev.name))
						subStatus = False
				if subStatus == False:
					self.socoUnsubscribe (dev)
					self.socoSubscribe (dev)


	def socoResubscribeAll(self):
		for dev in indigo.devices.iter("self.ZonePlayer"):
			if dev.enabled:
				self.socoUnsubscribe (dev)
				self.socoSubscribe (dev)		

	def socoResubscribe(self, dev, bootseq):
		self.plugin.errorLog("[%s] Network Error causing resubscription (bootseq=%s,%s) for ZonePlayer: %s" % (time.asctime(), dev.states["bootseq"], bootseq, dev.name))
		self.updateStateOnServer (dev, "bootseq", bootseq)
		self.socoUnsubscribe (dev)
		self.socoSubscribe (dev)

	def socoSubscribe(self, dev):
		self.soco_sub[dev.id] = {'soco_dev':SoCo(dev.address),
				'avTransport':SoCo(dev.address).avTransport.subscribe(requested_timeout=600, auto_renew=True), 
				'renderingControl':SoCo(dev.address).renderingControl.subscribe(requested_timeout=600, auto_renew=True), 
				'zoneGroupTopology':SoCo(dev.address).zoneGroupTopology.subscribe(requested_timeout=600, auto_renew=True), 
				'queue':SoCo(dev.address).queue.subscribe(requested_timeout=600, auto_renew=True), 
				'groupRenderingControl':SoCo(dev.address).groupRenderingControl.subscribe(requested_timeout=600, auto_renew=True), 
				'contentDirectory':SoCo(dev.address).contentDirectory.subscribe(requested_timeout=600, auto_renew=True), 
				'groupManagement':SoCo(dev.address).groupManagement.subscribe(requested_timeout=600, auto_renew=True)}
		if int(dev.pluginProps["model"]) in [SONOS_CONNECT, SONOS_CONNECTAMP, SONOS_PLAY5]:
			self.soco_sub[dev.id]['audioIn'] = SoCo(dev.address).audioIn.subscribe(requested_timeout=600, auto_renew=True)

		self.soco_sub[dev.id]['avTransport'].callback = self.socoEvents
		self.soco_sub[dev.id]['renderingControl'].callback = self.socoEvents
		self.soco_sub[dev.id]['zoneGroupTopology'].callback = self.socoEvents
		self.soco_sub[dev.id]['queue'].callback = self.socoEvents
		self.soco_sub[dev.id]['groupRenderingControl'].callback = self.socoEvents
		self.soco_sub[dev.id]['contentDirectory'].callback = self.socoEvents
		self.soco_sub[dev.id]['groupManagement'].callback = self.socoEvents
		if int(dev.pluginProps["model"]) in [SONOS_CONNECT, SONOS_CONNECTAMP, SONOS_PLAY5]:
			self.soco_sub[dev.id]['audioIn'].callback = self.socoEvents

	def socoUnsubscribe(self, dev):
		self.plugin.debugLog("[%s] Unsubscribing to ZonePlayer: %s" % (time.asctime(), dev.name))
		try: self.soco_sub[dev.id]['avTransport'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from avTransport events for ZonePlayer: %s" % dev.name)
		try: self.soco_sub[dev.id]['renderingControl'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from renderingControl events for ZonePlayer: %s" % dev.name)
		try: self.soco_sub[dev.id]['zoneGroupTopology'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from zoneGroupTopology events for ZonePlayer: %s" % dev.name)
		try: self.soco_sub[dev.id]['queue'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from queue events for ZonePlayer: %s" % dev.name)
		try: self.soco_sub[dev.id]['groupRenderingControl'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from groupRenderingControl events for ZonePlayer: %s" % dev.name)
		try: self.soco_sub[dev.id]['contentDirectory'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from contentDirectory events for ZonePlayer: %s" % dev.name)
		try: self.soco_sub[dev.id]['groupManagement'].unsubscribe()
		except: self.plugin.errorLog("Cannot unsubscribe from groupManagement events for ZonePlayer: %s" % dev.name)
		if int(dev.pluginProps["model"]) in [SONOS_CONNECT, SONOS_CONNECTAMP, SONOS_PLAY5]:
			try: self.soco_sub[dev.id]['audioIn'].unsubscribe()
			except: self.plugin.errorLog("Cannot unsubscribe from audioIn events for ZonePlayer: %s" % dev.name)

	def getModelName(self, dev):
		url = u"http://" + dev.pluginProps["address"] + ":1400/xml/device_description.xml"
		response = requests.get(url)
		if response.ok:
			root = ET.fromstring(response.content)
			ModelName = root.findtext('.//{urn:schemas-upnp-org:device-1-0}displayName')
			self.updateStateOnServer (dev, "ModelName", ModelName)
		else:
			self.plugin.errorLog("[%s] Cannot get ModelName for ZonePlayer: %s" % (time.asctime(), dev.name))


	def updateZoneTopology(self, dev):
		url = u"http://" + dev.pluginProps["address"] + ":1400/status/topology"
		response = requests.get(url)
		if response.ok:
			root = ET.fromstring(response.content)
			for ZonePlayer in root.findall("./ZonePlayers/ZonePlayer"):
				if ZonePlayer.get('uuid') == dev.states['ZP_LocalUID']:
					self.updateStateOnServer (dev, "GROUP_Coordinator", ZonePlayer.get('coordinator'))
					self.updateStateOnServer (dev, "bootseq", ZonePlayer.get('bootseq'))
		else:
			self.plugin.errorLog("[%s] Cannot get ZoneGroupTopology for ZonePlayer: %s" % (time.asctime(), dev.name))

	def updateZoneGroupStates(self, dev):
		zoneIP = dev.pluginProps["address"]
		res = self.SOAPSend (zoneIP, "/ZonePlayer", "/ZoneGroupTopology", "GetZoneGroupAttributes", "")
		self.updateStateOnServer (dev, "ZoneGroupName", self.parseCurrentZoneGroupName(res).decode('utf-8'))
		self.updateStateOnServer (dev, "ZoneGroupID", self.parseCurrentZoneGroupID(res))
		self.updateStateOnServer (dev, "ZonePlayerUUIDsInGroup", self.parseCurrentZonePlayerUUIDsInGroup(res))

	def IVONAVoices(self):
		global IVONAVoices
		id = 0
		IVONAjson = json.load(open('pyvona.json'))
		for majorkey, subdict in IVONAjson.iteritems():
			for line in subdict:
				#indigo.server.log("line: %s" % line)
				for subkey, value in line.iteritems():
					if subkey == 'Gender':
						IVONAgender = value
					elif subkey == 'Language':
						IVONAlanguage = value
					elif subkey == 'Name':
						IVONAname = value
				IVONAVoices.append([id, IVONAname, IVONAgender, IVONAlanguage, IVONAlanguages[IVONAlanguage]])
				self.plugin.debugLog("\tIVONA: %s|%s|%s" % (id, IVONAlanguages[IVONAlanguage], IVONAname))
				id = id + 1
		indigo.server.log(u"Loaded IVONA Voices... [%s]" % id)

	def PollyVoices(self):
		global PollyVoices
		id = 0
		client = boto3.client('polly', aws_access_key_id=self.PollyaccessKey, aws_secret_access_key=self.PollysecretKey, region_name='us-east-1')
		content = client.describe_voices()
		response = int(content["ResponseMetadata"]["HTTPStatusCode"])
		if response == 200:
			voices = content["Voices"]
			for voice in voices:
				PollyVoices.append([voice["Id"], voice["Name"], voice["Gender"], voice["LanguageCode"], voice["LanguageName"]])
				self.plugin.debugLog("\tPolly: %s|%s|%s" % (voice["Id"], voice["LanguageName"], voice["Name"]))
				id = id + 1
			indigo.server.log(u"Loaded Polly Voices... [%s]" % id)


	def SOAPSend(self, zoneIP, soapRoot, soapBranch, soapAction, soapPayload):
		if soapBranch == "/Queue":
			urn = "schemas-sonos-com"
		else:
			urn = "schemas-upnp-org"

		self.plugin.debugLog ("zoneIP: %s, soapRoot: %s, soapBranch: %s, soapAction: %s" % (zoneIP, soapRoot, soapBranch, soapAction))

		#SM_TEMPLATE = """<?xml version="1.0" encoding="utf-8"?><s:Envelope s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"><s:Body><ns0:%s xmlns:ns0=\"urn:%s:service:%s:1\"><InstanceID>0</InstanceID>%s</ns0:%s></s:Body></s:Envelope>""" % (soapAction, urn, soapBranch[1:], soapPayload, soapAction)
		
		SM_TEMPLATE = (
			'<?xml version="1.0" encoding="utf-8"?>'
			'<s:Envelope s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">'
				'<s:Body>'
					'<ns0:' + soapAction + ' xmlns:ns0="urn:' + urn + ':service:' + soapBranch[1:] + ':1">'
						'<InstanceID>0</InstanceID>'
						+ soapPayload +
					'</ns0:' + soapAction + '>'
				'</s:Body>'
			'</s:Envelope>')

		SoapMessage = SM_TEMPLATE

		base_url = 'http://' + zoneIP + ':1400'
		
		if soapRoot == "/ZonePlayer":
			control_url = soapBranch + '/Control'
		else:
			control_url = soapRoot + soapBranch + '/Control'

		soap_action = 'urn:' + urn + ':service:' + soapBranch[1:] + ':1#' + soapAction
		headers = {'Content-Type': 'text/xml; charset="utf-8"',	'Content-Length': str(len(SoapMessage)), 'Host': zoneIP + ':1400', 'User-Agent': 'Indigo', 'SOAPACTION': soap_action}

		try:
			response = requests.post(base_url + control_url, headers=headers, data=SoapMessage.encode("utf-8"))
		except Exception, e:
			self.plugin.errorLog("SOAPSend Error: %s" % e)

		res = response.text.encode("utf-8")
		status = response.status_code
		if status != 200:
			try:
				errorCode = self.parseErrorCode(res)
				self.plugin.errorLog("UPNP Error: %s" % UPNP_ERRORS[errorCode])
			except:
				self.plugin.errorLog("UPNP Error: %s" % status)
			self.plugin.errorLog ("Offending Command -> zoneIP: %s, soapRoot: %s, soapBranch: %s, soapAction: %s" % (zoneIP, soapRoot, soapBranch, soapAction))
			self.plugin.debugLog("Error Response: %s" % res)

		# reconstruct multi-line strings
		# discovered in Sonos beta firmware
		resx = ""

		for line in res.splitlines():
		##f = open ('beta.txt')
		##for line in iter(f):
			#indigo.server.log ("line: %s" % line)
			if len(line) <= 5:
				try:
					if int(line, 16) >= 0 and int(line,16) <= 4096:
						pass
					else:
						resx = resx + line.rstrip('\n')
				except:
					pass
			else:
				resx = resx + line.rstrip('\n')
					
		if self.plugin.xmlDebug:
			self.plugin.debugLog(SoapMessage)
			self.plugin.debugLog(resx)
		
		return (resx)

	def socoEvents (self, soco_event):
		deviceFound = False
		for deviceId in self.deviceList:
			if soco_event.service.soco.ip_address == indigo.devices[int(deviceId)].address:
				dev = indigo.devices[int(deviceId)]
				deviceFound = True
				break
		if deviceFound == False:
			self.plugin.errorLog("[%s] Cannot find Indigo device for ZonePlayer: %s" % (time.asctime(), soco_event.service.soco.ip_address))
			return

		service = soco_event.service.service_type

		# avTransport
		if service == "AVTransport":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo AVTransport] %s" % (time.asctime(), dev.name, soco_event.variables))

			# insures topology updates if AVTransport event received before Group event
			self.updateZoneGroupStates (dev)
			self.updateZoneTopology (dev)

			try:
				val = soco_event.variables["transport_state"]
				if val == "PAUSED_PLAYBACK":
					val = "PAUSED"
				self.updateStateOnServer (dev, "ZP_STATE", val)
			except: pass
			try:
				val = soco_event.variables["current_play_mode"]
				if val == "NORMAL":
					self.updateStateOnServer (dev, "Q_Repeat", "off")
					self.updateStateOnServer (dev, "Q_RepeatOne", "off")					
					self.updateStateOnServer (dev, "Q_Shuffle", "off")
				elif val == "REPEAT_ALL":
					self.updateStateOnServer (dev, "Q_Repeat", "on")
					self.updateStateOnServer (dev, "Q_RepeatOne", "off")
					self.updateStateOnServer (dev, "Q_Shuffle", "off")
				elif val == "REPEAT_ONE":
					self.updateStateOnServer (dev, "Q_Repeat", "off")
					self.updateStateOnServer (dev, "Q_RepeatOne", "on")
					self.updateStateOnServer (dev, "Q_Shuffle", "off")
				elif val == "SHUFFLE_NOREPEAT":
					self.updateStateOnServer (dev, "Q_Repeat", "off")
					self.updateStateOnServer (dev, "Q_RepeatOne", "off")
					self.updateStateOnServer (dev, "Q_Shuffle", "on")
				elif val == "SHUFFLE":
					self.updateStateOnServer (dev, "Q_Repeat", "on")
					self.updateStateOnServer (dev, "Q_RepeatOne", "off")
					self.updateStateOnServer (dev, "Q_Shuffle", "on")
				elif val == "SHUFFLE_REPEAT_ONE":
					self.updateStateOnServer (dev, "Q_Repeat", "off")
					self.updateStateOnServer (dev, "Q_RepeatOne", "on")
					self.updateStateOnServer (dev, "Q_Shuffle", "on")
			except: pass
			try:
				val = soco_event.variables["current_crossfade_mode"]
				if val == "0":
					self.updateStateOnServer (dev, "Q_Crossfade", "off")
				elif val == "1":
					self.updateStateOnServer (dev, "Q_Crossfade", "on")
			except: pass
			try: self.updateStateOnServer (dev, "ZP_CurrentTrack", soco_event.variables["current_track"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_CurrentTrackURI", soco_event.variables["current_track_uri"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_CurrentURI", soco_event.variables["av_transport_uri"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_DURATION", soco_event.variables["current_track_duration"])
			except: pass

			try:
				val = soco_event.variables["current_track_meta_data"]
				try:
					#if uri_radio not in val.title.decode('utf-8'):
					if uri_radio not in dev.states["ZP_CurrentURI"]:
						self.updateStateOnServer (dev, "ZP_TRACK", val.title)
				except:
					self.updateStateOnServer (dev, "ZP_TRACK", "")
				try: self.updateStateOnServer (dev, "ZP_CREATOR", val.creator)
				except: self.updateStateOnServer (dev, "ZP_CREATOR", "")
				try: self.updateStateOnServer (dev, "ZP_ALBUM", val.album)
				except: self.updateStateOnServer (dev, "ZP_ALBUM", "")
				try: self.updateStateOnServer (dev, "ZP_ARTIST", val.artist)
				except: self.updateStateOnServer (dev, "ZP_ARTIST", "")
				try:
					self.processArt (dev, val.album_art_uri.decode('utf-8'))
					if dev.states['GROUP_Coordinator'] == "true":
						ZonePlayerUUIDsInGroup = dev.states['ZonePlayerUUIDsInGroup'].split(',')
						for rdev in indigo.devices.iter("self.ZonePlayer"):
							SlaveUID = rdev.states['ZP_LocalUID']
							if SlaveUID != dev.states['ZP_LocalUID'] and SlaveUID in ZonePlayerUUIDsInGroup:
								self.processArt (rdev, val.album_art_uri)
				except:
					self.processArt (dev, None)
					#self.updateStateOnServer (dev, "ZP_ART", "")

				try:
					if uri_siriusxm in dev.states["ZP_CurrentURI"]:
						mediaInfo = val.stream_content.split("|")
						for item in mediaInfo:
							if "TITLE" in item:
								self.updateStateOnServer (dev, "ZP_TRACK", item[6:])
							elif "ARTIST" in item:
								self.updateStateOnServer (dev, "ZP_ARTIST", item[7:])
						self.updateStateOnServer (dev, "ZP_INFO", "")
					else:
						self.updateStateOnServer (dev, "ZP_INFO", val.stream_content)
				except:
					self.updateStateOnServer (dev, "ZP_INFO", "")

				# need to figure out when this is really necessary - can't populate for music library
				try: self.updateStateOnServer (dev, "ZP_STATION", soco_event.variables["enqueued_transport_uri_meta_data"].title)
				except: self.updateStateOnServer (dev, "ZP_STATION", "")

			except: pass

			try: 
				val = soco_event.variables["next_track_meta_data"]
				try: self.updateStateOnServer (dev, "ZP_NTRACK", val.title)
				except: self.updateStateOnServer (dev, "ZP_NTRACK", "")
				try: self.updateStateOnServer (dev, "ZP_NCREATOR", val.creator)
				except: self.updateStateOnServer (dev, "ZP_NCREATOR", "")
				try: self.updateStateOnServer (dev, "ZP_NALBUM", val.album)
				except: self.updateStateOnServer (dev, "ZP_NALBUM", "")
				try: self.updateStateOnServer (dev, "ZP_NARTIST", val.artist)
				except: self.updateStateOnServer (dev, "ZP_NARTIST", "")
				try: self.updateStateOnServer (dev, "ZP_NART", val.album_art_uri)
				except: self.updateStateOnServer (dev, "ZP_NART", "")
			except: pass
			
		# renderingControl
		elif service == "RenderingControl":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo RenderingControl] %s" % (time.asctime(), dev.name, soco_event.variables))

			try: self.updateStateOnServer (dev, "ZP_VOLUME", soco_event.variables["volume"]["Master"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_MUTE", soco_event.variables["mute"]["Master"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_VOLUME_FIXED", soco_event.variables["output_fixed"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_BASS", soco_event.variables["bass"])
			except: pass
			try: self.updateStateOnServer (dev, "ZP_TREBLE", soco_event.variables["treble"])
			except: pass

		# queue
		elif service == "Queue":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo Queue] %s" % (time.asctime(), dev.name, soco_event.variables))

			val = soco_event.variables["update_id"]
			if actionBusy == 0:
					PlaylistName = "Indigo_" + str(dev.states['ZP_LocalUID'])
					self.actionDirect (PA(dev.id, {"setting":PlaylistName}), "Q_Save")

		# groupRenderingControl
		elif service == "GroupRenderingControl":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo GroupRenderingControl] %s" % (time.asctime(), dev.name, soco_event.variables))

		# contentDiretory
		elif service == "ContentDirectory":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo ContentDirectory] %s" % (time.asctime(), dev.name, soco_event.variables))

			if dev.pluginProps['address'] == self.rootZPIP:
				try:
					val = soco_event.variables["container_update_i_ds"]
					if "SQ:" in val:
						self.getPlaylistsDirect()
					if "R:0" in val:
						self.getRT_FavStationsDirect()
					if "FV:2" in val:
						self.getSonosFavorites()
				except: pass

		# groupManagement
		elif service == "GroupManagement":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo GroupManagement] %s" % (time.asctime(), dev.name, soco_event.variables))

			try: self.updateStateOnServer (dev, "GROUP_Coordinator", str(bool(int(soco_event.variables["group_coordinator_is_local"]))).lower())
			except: pass
			try: self.updateStateOnServer (dev, "GROUP_Name", soco_event.variables["local_group_uuid"].decode('utf-8'))
			except: pass
			if dev.states["GROUP_Coordinator"] == False:
				self.copyStateFromMaster (dev)
			else:
				self.updateStateOnSlaves (dev)

		# zoneGroupTopology
		elif service == "ZoneGroupTopology":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo ZoneGroupTopology] %s" % (time.asctime(), dev.name, soco_event.variables))

			try: self.updateStateOnServer (dev, "ZoneGroupID", soco_event.variables["zone_group_id"].decode('utf-8'))
			except: pass
			try: self.updateStateOnServer (dev, "ZoneGroupName", soco_event.variables["zone_group_name"].decode('utf-8'))
			except: pass
			try: self.updateStateOnServer (dev, "ZonePlayerUUIDsInGroup", soco_event.variables["zone_player_uui_ds_in_group"].decode('utf-8'))
			except: pass
			if dev.states['ZoneGroupName'] in ["", "None"]:
				self.copyStateFromMaster (dev)
			else:
				self.updateStateOnSlaves (dev)
			"""
			try:
				ZGS = XML(soco_event.variables["zone_group_state"].decode('utf-8'))
				for elem in ZGS:
					for ZoneGroupMember in elem.findall("./ZoneGroupMember"):
						ZGM_UUID = ZoneGroupMember.get('UUID')
						ZGM_BootSeq = ZoneGroupMember.get('BootSeq')
						for deviceId in self.deviceList:
							dev = indigo.devices[int(deviceId)]
							if dev.states["ZP_LocalUID"] == ZGM_UUID:
								if int(ZGM_BootSeq) > int(dev.states["bootseq"]):
									self.socoResubscribe(dev, ZGM_BootSeq)
								break
			except:
				pass
			"""

		# audioIn
		elif service == "AudioIn":
			if self.plugin.eventsDebug:
				self.plugin.debugLog("[%s] [%s] [SoCo AudioIn] %s" % (time.asctime(), dev.name, soco_event.variables))

			val = soco_event.variables["audio_input_name"]
			self.updateStateOnServer (dev, "ZP_AIName", val.decode('utf-8'))
			self.getLineIn (dev, val.decode('utf-8'))

	def processArt (self, dev, val):
		self.plugin.debugLog("Processing Cover Art: %s:%s" % (dev.name, val))
		ZP_ZoneName = dev.states["ZP_ZoneName"]
		res = dev.states["ZP_CurrentURI"]
		if val != None:
			# Retrieve parent URI for slave zone
			if uri_group in res:
				(ParentUID,x) = dev.states["ZoneGroupID"].split(':')
				for pdev in indigo.devices.iter("self.ZonePlayer"):
					if ParentUID == pdev.states["ZP_LocalUID"]:
						res = pdev.states["ZP_CurrentURI"]
			prev_art = dev.states["ZP_ART"]
			if uri_radio in res:
				loc = str(res).find(uri_radio)
				if (loc >= 0):
					loc_beg = loc + len(uri_radio)
					loc_end = str(res).find("?sid", loc_beg)
					loc_end = str(res).find("?sid", loc_beg)
					val = "http://d1i6vahw24eb07.cloudfront.net/"+self.restoreString(str(res)[loc_beg:loc_end],0)+"q.png"
					self.updateStateOnServer (dev, "ZP_ART", val.decode('utf-8'))
					if val != prev_art:
						reqObj = urllib2.Request(val)
						fileObj = urllib2.urlopen(reqObj)
						localFile = open("/Library/Application Support/Perceptive Automation/images/Sonos/"+ZP_ZoneName+"_art.jpg", "wb")
						localFile.write(fileObj.read())
						localFile.close()
			elif uri_pandora in res or uri_file in res or uri_music or uri_playlist in res:
				if str(val).find("/getaa?") >= 0:
					val = "http://"+self.rootZPIP+":1400"+val
				self.updateStateOnServer (dev, "ZP_ART", val.decode('utf-8'))
				if val != prev_art:
					reqObj = urllib2.Request(val)
					fileObj = urllib2.urlopen(reqObj)
					localFile = open("/Library/Application Support/Perceptive Automation/images/Sonos/"+ZP_ZoneName+"_art.jpg", "wb")
					localFile.write(fileObj.read())
					localFile.close()
			elif uri_siriusxm in res:
				self.updateStateOnServer (dev, "ZP_ART", val.decode('utf-8'))
				if val != prev_art:
					try:
						reqObj = urllib2.Request(val)
						fileObj = urllib2.urlopen(reqObj)
						localFile = open("/Library/Application Support/Perceptive Automation/images/Sonos/"+ZP_ZoneName+"_art.jpg", "wb")
						localFile.write(fileObj.read())
						localFile.close()
					except:
						shutil.copy2("sonos_art.jpg", "/Library/Application Support/Perceptive Automation/images/Sonos/"+ZP_ZoneName+"_art.jpg")
			
		else:
			self.updateStateOnServer (dev, "ZP_ART", "")
			if uri_tv in res:
				try: shutil.copy2("sonos_tv.jpg", "/Library/Application Support/Perceptive Automation/images/Sonos/"+ZP_ZoneName+"_art.jpg")
				except:	pass
			else:
				try: shutil.copy2("sonos_art.jpg", "/Library/Application Support/Perceptive Automation/images/Sonos/"+ZP_ZoneName+"_art.jpg")
				except: pass

	def parsePoint (self, res, startString, stopString):
		loc = str(res).find(startString)
		if (loc > 0):
			loc_beg = loc + len(startString)
			loc_end = str(res).find(stopString, loc_beg)
			return (self.restoreString(str(res)[loc_beg:loc_end],0))
		else:
			return ""

	def parseDirty (self, res, startString, stopString):
		loc = str(res).find(startString)
		if (loc > 0):
			loc_beg = loc + len(startString)
			loc_end = str(res).find(stopString, loc_beg)
			return (str(res)[loc_beg:loc_end])
		else:
			return ""
	
	def parseFirstTrackNumberEnqueued(self, deviceId, res):
		loc = str(res).find("<FirstTrackNumberEnqueued>")
		if (loc > 0):
			loc_beg = loc + len("<FirstTrackNumberEnqueued>")
			loc_end = str(res).find("</FirstTrackNumberEnqueued>", loc_beg)
			item = self.restoreString(str(res)[loc_beg:loc_end],0)
			return item

	def parseRelTime(self, deviceId, res):
		return self.parsePoint (res, "<RelTime>", "</RelTime>")

	def parseCurrentZoneGroupName(self, res):
		return self.parsePoint (res, "<CurrentZoneGroupName>", "</CurrentZoneGroupName>")

	def parseCurrentZoneGroupID(self, res):
		return self.parsePoint (res, "<CurrentZoneGroupID>", "</CurrentZoneGroupID>")

	def parseCurrentZonePlayerUUIDsInGroup(self, res):
		return self.parsePoint (res, "<CurrentZonePlayerUUIDsInGroup>", "</CurrentZonePlayerUUIDsInGroup>")

	def parseCurrentVolume(self, res):
		return self.parsePoint (res, "<CurrentVolume>", "</CurrentVolume>")

	def parseCurrentMute(self, res):
		return self.parsePoint (res, "<CurrentMute>", "</CurrentMute>")

	def parseCurrentTransportActions(self, res):
		return self.parsePoint (res, "<Actions>", "</Actions>")

	def parseErrorCode(self, res):
		return self.parsePoint (res, "<errorCode>", "</errorCode>")

	def parseBrowseNumberReturned(self, res):
		return self.parsePoint (res, "<NumberReturned>", "</NumberReturned>")

	def parseAssignedObjectID(self, res):
		return self.parsePoint (res, "<AssignedObjectID>", "</AssignedObjectID>")

	def parsePandoraToken(self, res):
		return self.parsePoint (res, "&m=", "&f")

	def playRadio(self, zoneIP, l2p):
		self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>"+l2p+"</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"-1\" parentID=\"-1\" restricted=\"true\"&gt;&lt;dc:title&gt;RADIO&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON65031_&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
		self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")

	def playPandora(self, zoneIP, l2p, pstation):
		for pandora_station in Sonos_Pandora:
			if pandora_station[0] == l2p:
				pemail = pandora_station[2]
				break		
		self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>pndrradio:"+l2p+"</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"OOOX" + l2p + "\" parentID=\"0\" restricted=\"true\"&gt;&lt;dc:title&gt;" + pstation +  "&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON3_" + pemail + "&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
		self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")

	def boolConv(self, val):
		if val == "on":
			return True
		else:
			return False

	def QMode(self, repeat, repeat_one, shuffle):
		if (repeat == False and repeat_one == False and shuffle == False):
			PlayMode = "NORMAL"
		elif (repeat == True and shuffle == False):
			PlayMode = "REPEAT_ALL"
		elif (repeat_one == True and shuffle == False):
			PlayMode = "REPEAT_ONE"
		elif (repeat == False and repeat_one == False and shuffle == True):
			PlayMode = "SHUFFLE_NOREPEAT"
		elif (repeat == True and shuffle == True):
			PlayMode = "SHUFFLE"
		elif (repeat_one == True and shuffle == True):
			PlayMode = "SHUFFLE_REPEAT_ONE"
		return PlayMode

	######################################################################################
	# Actions

	def actionDirect(self, pluginAction, action):
		global Sonos_Playlists
		global Sonos_Pandora
		global Sonos_RT_FavStations

		if action != "setStandalones":

			dev = indigo.devices[pluginAction.deviceId]
			zoneIP = dev.pluginProps["address"]

			if dev.states["GROUP_Coordinator"] == "false":
				Coordinator = dev.states["GROUP_Name"]
				for idev in indigo.devices.iter("self.ZonePlayer"):
					if idev.states["GROUP_Coordinator"] == "true" and idev.states["GROUP_Name"] == Coordinator:
						CoordinatorIP = idev.pluginProps["address"]
						CoordinatorDev = idev
						break

		if action == "Play":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			self.plugin.debugLog("Sonos Action: Play")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			indigo.server.log("ZonePlayer: %s, Play" % dev.name)
		elif action == "TogglePlay":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			self.plugin.debugLog("Sonos Action: Toggle Play")
			if (dev.states["ZP_STATE"] == "PLAYING"):
				self.actionDirect(pluginAction, "Pause")
				indigo.server.log("ZonePlayer: %s, Pause" % dev.name)
			else:
				self.actionDirect(pluginAction, "Play")
				indigo.server.log("ZonePlayer: %s, Play" % dev.name)
		elif action == "Pause":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			self.plugin.debugLog("Sonos Action: Pause")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Pause", "")
			indigo.server.log("ZonePlayer: %s, Pause" % dev.name)
		elif action == "Stop":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			self.plugin.debugLog("Sonos Action: Stop")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Stop", "")
			indigo.server.log("ZonePlayer: %s, Stop" % dev.name)
		elif action == "Previous":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			self.plugin.debugLog("Sonos Action: Previous")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Previous", "")
			indigo.server.log("ZonePlayer: %s, Previous Track" % dev.name)
		elif action == "Next":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			self.plugin.debugLog("Sonos Action: Next")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Next", "")
			indigo.server.log("ZonePlayer: %s, Next Track" % dev.name)

		elif action == "MuteToggle":
			self.plugin.debugLog("Sonos Action: Mute Toggle")
			if int(dev.states["ZP_MUTE"]) == 0:
				self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>1</DesiredMute>")
				indigo.server.log("ZonePlayer: %s, Mute On" % dev.name)
			else:
				self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>0</DesiredMute>")
				indigo.server.log("ZonePlayer: %s, Mute Off" % dev.name)
		elif action == "MuteOn":
			self.plugin.debugLog("Sonos Action: Mute On")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>1</DesiredMute>")
			indigo.server.log("ZonePlayer: %s, Mute On" % dev.name)
		elif action == "MuteOff":
			self.plugin.debugLog("Sonos Action: Mute Off")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>0</DesiredMute>")
			indigo.server.log("ZonePlayer: %s, Mute Off" % dev.name)

		elif action == "GroupMuteToggle":
			self.plugin.debugLog("Sonos Action: Group Mute Toggle")
			if int(self.parseCurrentMute (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupMute", ""))) == 0:
				self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetGroupMute", "<DesiredMute>1</DesiredMute>")
				indigo.server.log("ZonePlayer Group: %s, Mute On" % dev.name)
			else:
				self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetGroupMute", "<DesiredMute>0</DesiredMute>")
				indigo.server.log("ZonePlayer Group: %s, Mute Off" % dev.name)
		elif action == "GroupMuteOn":
			self.plugin.debugLog("Sonos Action: Group Mute On")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetGroupMute", "<DesiredMute>1</DesiredMute>")
			indigo.server.log("ZonePlayer Group: %s, Mute On" % dev.name)
		elif action == "GroupMuteOff":
			self.plugin.debugLog("Sonos Action: Group Mute Off")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetGroupMute", "<DesiredMute>0</DesiredMute>")
			indigo.server.log("ZonePlayer Group: %s, Mute Off" % dev.name)

		elif action == "Volume":
			self.plugin.debugLog("Sonos Action: Volume")
			current_volume = dev.states["ZP_VOLUME"]
			new_volume = pluginAction.props.get("setting")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetVolume", "<Channel>Master</Channel><DesiredVolume>"+str(new_volume)+"</DesiredVolume>")
			indigo.server.log(u"ZonePlayer: %s, Current Volume: %s, New Volume: %s" % (dev.name, current_volume, new_volume))
		elif action == "VolumeDown":
			self.plugin.debugLog("Sonos Action: Volume Down")
			current_volume = dev.states["ZP_VOLUME"]
			new_volume = int(current_volume) - 2
			if new_volume < 0:
				new_volume = 0
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetVolume", "<Channel>Master</Channel><DesiredVolume>"+str(new_volume)+"</DesiredVolume>")
			indigo.server.log(u"ZonePlayer: %s, Current Volume: %s, New Volume: %s" % (dev.name, current_volume, new_volume))
		elif action == "VolumeUp":
			self.plugin.debugLog("Sonos Action: Volume Up")
			current_volume = dev.states["ZP_VOLUME"]
			new_volume = int(current_volume) + 2
			if new_volume > 100:
				new_volume = 100
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetVolume", "<Channel>Master</Channel><DesiredVolume>"+str(new_volume)+"</DesiredVolume>")
			indigo.server.log(u"ZonePlayer: %s, Current Volume: %s, New Volume: %s" % (dev.name, current_volume, new_volume))

		elif action == "GroupVolume":
			self.plugin.debugLog("Sonos Action: Group Volume")
			current_volume = self.parseCurrentVolume (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupVolume", ""))
			new_volume = pluginAction.props.get("setting")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetGroupVolume", "<DesiredVolume>"+str(new_volume)+"</DesiredVolume>")
			indigo.server.log(u"ZonePlayer Group: %s, Current Group Volume: %s, New Group Volume: %s" % (dev.name, current_volume, new_volume))
		elif action == "RelativeGroupVolume":
			self.plugin.debugLog("Sonos Action: Relative Group Volume")
			current_volume = self.parseCurrentVolume (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupVolume", ""))
			adjustment = pluginAction.props.get("setting")
			new_volume = int(current_volume) + int(adjustment)
			if new_volume < 0:
				new_volume = 0
			if new_volume > 100:
				new_volume = 100
			self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetRelativeGroupVolume", "<Adjustment>"+adjustment+"</Adjustment>")
			indigo.server.log(u"ZonePlayer Group: %s, Current Group Volume: %s, New Group Volume: %s" % (dev.name, current_volume, new_volume))
		elif action == "GroupVolumeDown":
			self.plugin.debugLog("Sonos Action: Group Volume Down")
			current_volume = self.parseCurrentVolume (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupVolume", ""))
			new_volume = int(current_volume) - 2
			if new_volume < 0:
				new_volume = 0
			self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetRelativeGroupVolume", "<Adjustment>-2</Adjustment>")
			indigo.server.log(u"ZonePlayer Group: %s, Current Group Volume: %s, New Group Volume: %s" % (dev.name, current_volume, new_volume))
		elif action == "GroupVolumeUp":
			self.plugin.debugLog("Sonos Action: Group Volume Up")
			current_volume = self.parseCurrentVolume (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupVolume", ""))
			new_volume = int(current_volume) + 2
			if new_volume > 100:
				new_volume = 100
			self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetRelativeGroupVolume", "<Adjustment>2</Adjustment>")
			indigo.server.log(u"ZonePlayer Group: %s, Current Group Volume: %s, New Group Volume: %s" % (dev.name, current_volume, new_volume))

		elif action == "Bass":
			self.plugin.debugLog("Sonos Action: Bass")
			current_bass = dev.states["ZP_BASS"]
			new_bass = pluginAction.props.get("setting")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetBass", "<DesiredBass>"+str(new_bass)+"</DesiredBass>")
			indigo.server.log(u"ZonePlayer: %s, Current Bass: %s, New Bass: %s" % (dev.name, current_bass, new_bass))
		elif action == "BassDown":
			self.plugin.debugLog("Sonos Action: Bass Down")
			current_bass = dev.states["ZP_BASS"]
			new_bass = int(current_bass) - 1
			if new_bass < -10:
				new_bass = -10
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetBass", "<DesiredBass>"+str(new_bass)+"</DesiredBass>")
			indigo.server.log(u"ZonePlayer: %s, Current Bass: %s, New Bass: %s" % (dev.name, current_bass, new_bass))
		elif action == "BassUp":
			self.plugin.debugLog("Sonos Action: Bass Up")
			current_bass = dev.states["ZP_BASS"]
			new_volume = int(current_bass) + 1
			if new_bass > 10:
				new_bass = 10
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetBass", "<DesiredBass>"+str(new_bass)+"</DesiredBass>")
			indigo.server.log(u"ZonePlayer: %s, Current Bass: %s, New Bass: %s" % (dev.name, current_bass, new_bass))

		elif action == "Treble":
			self.plugin.debugLog("Sonos Action: Treble")
			current_treble = dev.states["ZP_BASS"]
			new_treble = pluginAction.props.get("setting")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetTreble", "<DesiredTreble>"+str(new_treble)+"</DesiredTreble>")
			indigo.server.log(u"ZonePlayer: %s, Current Treble: %s, New Treble: %s" % (dev.name, current_treble, new_treble))
		elif action == "TrebleDown":
			self.plugin.debugLog("Sonos Action: Treble Down")
			current_treble = dev.states["ZP_BASS"]
			new_treble = int(current_treble) - 1
			if new_treble < -10:
				new_treble = -10
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetTreble", "<DesiredTreble>"+str(new_treble)+"</DesiredTreble>")
			indigo.server.log(u"ZonePlayer: %s, Current Treble: %s, New Treble: %s" % (dev.name, current_treble, new_treble))
		elif action == "TrebleUp":
			self.plugin.debugLog("Sonos Action: Treble Up")
			current_treble = dev.states["ZP_BASS"]
			new_volume = int(current_treble) + 1
			if new_treble > 10:
				new_treble = 10
			self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetTreble", "<DesiredTreble>"+str(new_treble)+"</DesiredTreble>")
			indigo.server.log(u"ZonePlayer: %s, Current Treble: %s, New Treble: %s" % (dev.name, current_treble, new_treble))

		elif action == "Q_Crossfade":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			mode = pluginAction.props.get("setting")
			if mode == 0:
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetCrossfadeMode", "<CrossfadeMode>0</CrossfadeMode>")
			elif mode == 1:
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetCrossfadeMode", "<CrossfadeMode>1</CrossfadeMode>")
		elif action == "Q_Repeat":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			repeat = bool(int(pluginAction.props.get("setting")))
			repeat_one = self.boolConv(dev.states["Q_RepeatOne"])
			shuffle = self.boolConv(dev.states["Q_Shuffle"])
			if repeat == True:
				PlayMode = self.QMode(repeat, False, shuffle)
			else:
				PlayMode = self.QMode(repeat, repeat_one, shuffle)
			self.plugin.debugLog("Sonos Action: PlayMode %s" % PlayMode)
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetPlayMode", "<NewPlayMode>"+PlayMode+"</NewPlayMode>")
		elif action == "Q_RepeatOne":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			repeat_one = bool(int(pluginAction.props.get("setting")))
			repeat = self.boolConv(dev.states["Q_Repeat"])
			shuffle = self.boolConv(dev.states["Q_Shuffle"])
			if repeat_one == True:
				PlayMode = self.QMode(False, repeat_one, shuffle)
			else:
				PlayMode = self.QMode(repeat, repeat_one, shuffle)
			self.plugin.debugLog("Sonos Action: PlayMode %s" % PlayMode)
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetPlayMode", "<NewPlayMode>"+PlayMode+"</NewPlayMode>")
		elif action == "Q_RepeatToggle":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			repeat = self.boolConv(dev.states["Q_Repeat"])
			repeat_one = self.boolConv(dev.states["Q_RepeatOne"])
			shuffle = self.boolConv(dev.states["Q_Shuffle"])
			if repeat == False and repeat_one == False:
				PlayMode = self.QMode(True, False, shuffle)
			elif repeat == True and repeat_one == False:
				PlayMode = self.QMode(False, True, shuffle)
			else:
				PlayMode = self.QMode(False, False, shuffle)
			self.plugin.debugLog("Sonos Action: PlayMode %s" % PlayMode)
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetPlayMode", "<NewPlayMode>"+PlayMode+"</NewPlayMode>")
		elif action == "Q_Shuffle":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			shuffle = bool(int(pluginAction.props.get("setting")))
			repeat = self.boolConv(dev.states["Q_Repeat"])
			repeat_one = self.boolConv(dev.states["Q_RepeatOne"])
			PlayMode = self.QMode(repeat, repeat_one, shuffle)
			self.plugin.debugLog("Sonos Action: PlayMode %s" % PlayMode)
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetPlayMode", "<NewPlayMode>"+PlayMode+"</NewPlayMode>")
		elif action == "Q_ShuffleToggle":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			repeat = self.boolConv(dev.states["Q_Repeat"])
			repeat_one = self.boolConv(dev.states["Q_RepeatOne"])
			shuffle = self.boolConv(dev.states["Q_Shuffle"])
			if shuffle == True:
				PlayMode = self.QMode(repeat, repeat_one, False)
			else:
				PlayMode = self.QMode(repeat, repeat_one, True)
			self.plugin.debugLog("Sonos Action: PlayMode %s" % PlayMode)
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetPlayMode", "<NewPlayMode>"+PlayMode+"</NewPlayMode>")
		elif action == "Q_Clear":
			self.SOAPSend (zoneIP, "/MediaRenderer", "/Queue", "RemoveAllTracks", "<QueueID>0</QueueID><UpdateID>0</UpdateID>")
			indigo.server.log("ZonePlayer: %s, Clear Queue" % dev.name)

		elif action == "Q_Save":
			self.updateZoneTopology(dev)
			if dev.states["GROUP_Coordinator"] == "false":
				indigo.server.log("ZonePlayer: %s, Cannot Save Queue for Slave" % dev.name)
			else:
				self.plugin.sleep(0.5)
				PlaylistName = pluginAction.props.get("setting")
				ZP  = self.parseBrowseNumberReturned(self.SOAPSend (zoneIP, "/MediaServer", "/ContentDirectory", "Browse", "<ObjectID>Q:0</ObjectID><BrowseFlag>BrowseDirectChildren</BrowseFlag><Filter></Filter><StartingIndex>0</StartingIndex><RequestedCount>1000</RequestedCount><SortCriteria></SortCriteria>"))
				if PlaylistName == "Indigo_" + dev.states['ZP_LocalUID']:
					self.updateStateOnServer (dev, "Q_Number", ZP)
				if int(ZP) > 0:
					ObjectID = ""
					for plist in Sonos_Playlists:
						if plist[1] == PlaylistName:
							ObjectID = plist[2]
					AssignedObjectID = self.parseAssignedObjectID(self.SOAPSend (zoneIP, "/MediaRenderer", "/Queue", "SaveAsSonosPlaylist", "<QueueID>0</QueueID><Title>" + PlaylistName + "</Title><ObjectID>" + ObjectID + "</ObjectID>"))
					if ObjectID == "":
						ObjectID = AssignedObjectID
					if PlaylistName.find(dev.states['ZP_LocalUID']) > -1:
						self.updateStateOnServer (dev, "Q_ObjectID", ObjectID)

					indigo.server.log ("ZonePlayer: %s, Save Queue: %s" % (dev.name, PlaylistName))
				else:
					if PlaylistName == "Indigo_" + dev.states['ZP_LocalUID']:
						ObjectID = ""
						for plist in Sonos_Playlists:
							if plist[1] == PlaylistName:
								ObjectID = plist[2]
								self.actionDirect(PA(dev.id, {"setting":ObjectID}), "CD_RemovePlaylist")
						self.updateStateOnServer (dev, "Q_ObjectID", "")
					indigo.server.log ("ZonePlayer: %s, Nothing in Queue to Save" % dev.name)

		elif action == "CD_RemovePlaylist":
			ObjectID = pluginAction.props.get("setting")
			for plist in Sonos_Playlists:
				if plist[2] == ObjectID:
					PlaylistName = plist[1]
					self.SOAPSend (zoneIP, "/MediaServer", "/ContentDirectory", "DestroyObject", "<ObjectID>" + ObjectID + "</ObjectID>")
					indigo.server.log ("ZonePlayer: %s, Remove Playlist: %s" % (dev.name, PlaylistName))

		elif action == "ChannelUp":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
				currentURI = idev.states["ZP_CurrentURI"]
			else:
				currentURI = dev.states["ZP_CurrentURI"]
			x = 0
			foundit = "false"
			if uri_radio in currentURI:
				for l2p in Sonos_RT_FavStations:
					if currentURI == l2p[0]:
						foundit = "true"
						break
					else:
						x = x + 1
				if foundit == "false":
					x = 0
				elif foundit == "true":
					if x == 0:
						x = len(Sonos_RT_FavStations) - 1
					else:
						x = x - 1
				l2p = Sonos_RT_FavStations[x][0].replace("&", "&amp;")
				self.playRadio (zoneIP, l2p)
			elif uri_pandora in currentURI:
				Sonos_Pandora_Sort = sorted(Sonos_Pandora, key=lambda a:a[1])
				indigo.server.log("%s" % Sonos_Pandora_Sort)
				for l2p in Sonos_Pandora_Sort:
					if currentURI[10:] == l2p[0]:
						foundit = "true"
						break
					else:
						x = x + 1
				if foundit == "false":
					x = 0
				elif foundit == "true":
					if x == 0:
						x = len(Sonos_Pandora_Sort) - 1
					else:
						x = x - 1
				l2p = Sonos_Pandora_Sort[x][0]
				pstation = self.cleanString(Sonos_Pandora_Sort[x][1])
				self.playPandora (zoneIP, l2p, pstation)
			elif uri_siriusxm in currentURI:
				#setting = currentURI[currentURI.find("%3a")+3:currentURI.find("?")]
				setting = urllib.unquote(currentURI[currentURI.find(":")+1:currentURI.find("?")])
				for Channel in range(len(Sonos_SiriusXM)):
					#if Sonos_SiriusXM[Channel][3] == setting:
					if Sonos_SiriusXM[Channel][1] == setting:
						if Channel + 1 < len(Sonos_SiriusXM):
							#new_setting = Sonos_SiriusXM[Channel+1][3]
							new_setting = Sonos_SiriusXM[Channel+1][1]
						else:
							#new_setting = Sonos_SiriusXM[0][3]
							new_setting = Sonos_SiriusXM[0][1]
						break
				self.actionDirect (PA(dev.id, {"setting":new_setting}), "ZP_SiriusXM")

		elif action == "ChannelDown":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
				currentURI = idev.states["ZP_CurrentURI"]
			else:
				currentURI = dev.states["ZP_CurrentURI"]
			x = 0
			foundit = "false"
			if uri_radio in currentURI:
				for l2p in Sonos_RT_FavStations:
					if currentURI == l2p[0]:
						foundit = "true"
						break
					else:
						x = x + 1
				if foundit == "false":
					x = 0
				elif foundit == "true":
					if x == (len(Sonos_RT_FavStations) - 1):
						x = 0
					else:
						x = x + 1
				l2p = Sonos_RT_FavStations[x][0].replace("&", "&amp;")
				self.playRadio (zoneIP, l2p)
			elif uri_pandora in currentURI:
				Sonos_Pandora_Sort = sorted(Sonos_Pandora, key=lambda a:a[1])
				indigo.server.log("%s" % Sonos_Pandora_Sort)
				for l2p in Sonos_Pandora_Sort:
					if currentURI[10:] == l2p[0]:
						foundit = "true"
						break
					else:
						x = x + 1
				if foundit == "false":
					x = 0
				elif foundit == "true":
					if x == (len(Sonos_Pandora_Sort) - 1):
						x = 0
					else:
						x = x + 1
				l2p = Sonos_Pandora_Sort[x][0]
				pstation = self.cleanString(Sonos_Pandora_Sort[x][1])
				self.playPandora (zoneIP, l2p, pstation)
			elif uri_siriusxm in currentURI:
				#setting = currentURI[currentURI.find("%3a")+3:currentURI.find("?")]
				setting = urllib.unquote(currentURI[currentURI.find(":")+1:currentURI.find("?")])
				for Channel in range(len(Sonos_SiriusXM)):
					#if Sonos_SiriusXM[Channel][3] == setting:
					if Sonos_SiriusXM[Channel][1] == setting:
						if Channel > 0:
							#new_setting = Sonos_SiriusXM[Channel-1][3]
							new_setting = Sonos_SiriusXM[Channel-1][1]
						else:
							#new_setting = Sonos_SiriusXM[len(Sonos_SiriusXM)-1][3]
							new_setting = Sonos_SiriusXM[len(Sonos_SiriusXM)-1][1]
						break
				self.actionDirect (PA(dev.id, {"setting":new_setting}), "ZP_SiriusXM")

		elif action == "ZP_LineIn":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			dev_src = indigo.devices[int(pluginAction.props.get("setting"))]
			self.plugin.debugLog(u"Playing LineIn: %s..." % dev_src.states["ZP_AIName"])
			dev_src_LocalUID = dev_src.states['ZP_LocalUID']
			if dev_src.states["ZP_AIName"] != "":
				self.plugin.debugLog("Sonos Action: Line-In")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-stream:"+str(dev_src_LocalUID)+"</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")

		elif action == "ZP_RT_FavStation":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			l2p = pluginAction.props.get("setting").replace("&", "&amp;")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>"+l2p+"</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"-1\" parentID=\"-1\" restricted=\"true\"&gt;&lt;dc:title&gt;RADIO&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON65031_&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")

		elif action == "ZP_LIST":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
				dev_src_LocalUID = CoordinatorDev.states['ZP_LocalUID']
			else:
				dev_src_LocalUID = dev.states['ZP_LocalUID']				
			l2p = pluginAction.props.get("setting")
			mode = pluginAction.props.get("mode")
			for plist in Sonos_Playlists:
				if plist[0] == l2p:
					plist_name = plist[1]
			#indigo.server.log(u"Queueing Playlist: %s: %s..." % (mode, plist_name))
			if mode == "Play Now":
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev_src_LocalUID)+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData></EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Seek", "<Unit>TRACK_NR</Unit><Target>"+track_pos+"</Target>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			elif mode == "Play Next":
				#current_track = self.parseCurrentTrack(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "GetPositionInfo", ""))
				current_track = dev.states['ZP_CurrentTrack']
				indigo.server.log(current_track)
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData></EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>"+str(int(current_track)+1)+"</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
			elif mode == "Add To Queue":
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData></EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
			elif mode == "Replace Queue":
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev_src_LocalUID)+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "RemoveAllTracksFromQueue", "")
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData></EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Seek", "<Unit>TRACK_NR</Unit><Target>"+track_pos+"</Target>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			indigo.server.log ("ZonePlayer: %s, Play Sonos Playlist: %s" % (dev.name, plist_name))

		elif action == "ZP_Queue":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
				dev_src_LocalUID = CoordinatorDev.states['ZP_LocalUID']
			else:
				dev_src_LocalUID = dev.states['ZP_LocalUID']
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev_src_LocalUID)+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
		elif action == "addPlayerToZone":
			dev_dest = indigo.devices[int(pluginAction.props.get("setting"))]
			#self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon:"+str(dev_dest.states['ZP_LocalUID'])+"</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
			self.SOAPSend (dev_dest.pluginProps["address"], "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon:"+str(dev.states['ZP_LocalUID'])+"</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
		elif action == "setStandalone":
			indigo.server.log("remove zone from group: %s" % dev.name)
			if dev.states['GROUP_Coordinator'] == "true":
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "BecomeCoordinatorOfStandaloneGroup", "")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev.states['ZP_LocalUID'])+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
		elif action == "addPlayersToZone":
			zones = []
			x = 1
			while x <= 12:
				ivar = 'zp' + str(x)
				if pluginAction.props.get(ivar) not in ["", None, "00000"]: 
					zones.append(pluginAction.props.get(ivar))
				x = x + 1

			for item in zones:
				indigo.server.log("add zone to group: %s" % item)
				dev_dest = indigo.devices[int(item)]
				self.SOAPSend (dev_dest.pluginProps["address"], "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon:"+str(dev.states['ZP_LocalUID'])+"</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")

		elif action == "setStandalones":
			zones = []
			x = 1
			while x <= 12:
				ivar = 'zp' + str(x)
				if pluginAction.props.get(ivar) not in ["", None, "00000"]: 
					zones.append(pluginAction.props.get(ivar))
				x = x + 1
			
			for item in zones:
				indigo.server.log("remove zone from group: %s" % item)
				dev = indigo.devices[int(item)]
				if dev.states['GROUP_Coordinator'] == "true":
					self.SOAPSend (dev.pluginProps["address"], "/MediaRenderer", "/AVTransport", "BecomeCoordinatorOfStandaloneGroup", "")
				self.SOAPSend (dev.pluginProps["address"], "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev.states['ZP_LocalUID'])+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")

		elif action == "ZP_sleepTimer":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			l2p = pluginAction.props.get("setting")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "ConfigureSleepTimer", "<NewSleepTimerDuration>" + l2p + "</NewSleepTimerDuration>")

		elif action == "ZP_Pandora":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			l2p = pluginAction.props.get("setting")
			for pandora_station in Sonos_Pandora:
				if pandora_station[0] == l2p:
					pstation = self.cleanString(pandora_station[1])
					pemail = pandora_station[2]
					pnickname = pandora_station[3]
					break
			#self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>pndrradio:"+l2p+"</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"OOOX" + l2p + "\" parentID=\"0\" restricted=\"true\"&gt;&lt;dc:title&gt;" + pstation +  "&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON3_" + pemail + "&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-sonosapi-radio:ST%3a"+l2p+"?sid=236&amp;flags=8300&amp;sn=10</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"OOOX" + l2p + "\" parentID=\"0\" restricted=\"true\"&gt;&lt;dc:title&gt;" + pstation +  "&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON3_" + pemail + "&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			indigo.server.log ("ZonePlayer: %s, Play Pandora: %s: %s" % (dev.name, pnickname, pstation))

		elif action == "ZP_SiriusXM":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			contentId = pluginAction.props.get("setting")
			
			for Channel in Sonos_SiriusXM:
				#if Channel[3] == contentId:
				if Channel[1] == contentId:
					#siriusChannelNo = Channel[0]
					#name = self.cleanString(Channel[4]).encode('ascii', 'xmlcharrefreplace')
					title = Channel[2]
					break
			#self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-sonosapi-hls:r%3a" + contentId + "?sid=37&amp;flags=8480&amp;sn=8</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"" + contentId + "\" parentID=\"0\" restricted=\"true\"&gt;&lt;dc:title&gt;" + str(siriusChannelNo) + " - " + name +  "&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON6_&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-sonosapi-hls:" + urllib.quote(contentId) + "?sid=37&amp;flags=8480&amp;sn=8</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"" + contentId + "\" parentID=\"0\" restricted=\"true\"&gt;&lt;dc:title&gt;" + title +  "&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioBroadcast&lt;/upnp:class&gt;&lt;desc id=\"cdudn\" nameSpace=\"urn:schemas-rinconnetworks-com:metadata-1-0/\"&gt;SA_RINCON6_&lt;/desc&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			indigo.server.log ("ZonePlayer: %s, Play SiriusXM: %s" % (dev.name, title))

		elif action == "ZP_Container":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
				dev_src_LocalUID = CoordinatorDev.states['ZP_LocalUID']
			else:
				dev_src_LocalUID = dev.states['ZP_LocalUID']				
			l2p = pluginAction.props.get("setting")
			mode = pluginAction.props.get("mode")
			#(uri_header, uri_detail) = l2p.split(':')
			for title in Sonos_Favorites:
				if title[0] == l2p:
					pTitle = self.cleanString(title[1]).encode('ascii', 'xmlcharrefreplace')
					MD = title[2]
					break
			if mode == "Play Now":
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev_src_LocalUID)+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData>"+MD+"</EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Seek", "<Unit>TRACK_NR</Unit><Target>"+track_pos+"</Target>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			elif mode == "Play Next":
				#current_track = self.parseCurrentTrack(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "GetPositionInfo", ""))
				current_track = dev.states['ZP_CurrentTrack']
				indigo.server.log(current_track)
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData>"+MD+"</EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>"+str(int(current_track)+1)+"</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
			elif mode == "Add To Queue":
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData>"+MD+"</EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
			elif mode == "Replace Queue":
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-queue:"+str(dev_src_LocalUID)+"#0</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "RemoveAllTracksFromQueue", "")
				track_pos = self.parseFirstTrackNumberEnqueued(dev, self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData>"+MD+"</EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>"))
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Seek", "<Unit>TRACK_NR</Unit><Target>"+track_pos+"</Target>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			indigo.server.log ("ZonePlayer: %s, Play: %s" % (dev.name, pTitle))

		elif action =="ZP_SonosRadio":
			if dev.states["GROUP_Coordinator"] == "false":
				zoneIP = CoordinatorIP
			l2p = pluginAction.props.get("setting")
			for title in Sonos_Favorites:
				if title[0] == l2p:
					pTitle = self.cleanString(title[1]).encode('ascii', 'xmlcharrefreplace')
					URI = title[3]
					MD = title[2]
					break
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>"+URI+"</CurrentURI><CurrentURIMetaData>"+MD+"</CurrentURIMetaData>")
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
			indigo.server.log ("ZonePlayer: %s, Play Radio: %s" % (dev.name, pTitle))

		elif action == "ZP_TV":
			self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-sonos-htastream:"+str(dev.states['ZP_LocalUID'])+":spdif</CurrentURI><CurrentURIMetaData>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns:r=\"urn:schemas-rinconnetworks-com:metadata-1-0/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;&lt;item id=\"spdif-input\" parentID=\"0\" restricted=\"false\"&gt;&lt;dc:title&gt;"+str(dev.states['ZP_LocalUID'])+"&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.audioItem.audioItem&lt;/upnp:class&gt;&lt;res protocolInfo=\"spdif\"&gt;x-sonos-htastream:"+str(dev.states['ZP_LocalUID'])+":spdif&lt;/res&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</CurrentURIMetaData>")

		elif action == "ZP_SonosFavorites":
			setting = pluginAction.props.get("setting")
			for uri in Sonos_Favorites:
				if uri[4] == setting:
					l2p=uri[0]
					break
			mode = pluginAction.props.get("mode")
			if mode == "":
				mode = "Play Now"

			if uri_radio in l2p:
				self.actionDirect (PA(dev.id, {"setting":l2p}), "ZP_RT_FavStation")
			elif uri_pandora in l2p:
				setting = l2p[l2p.find(":")+1:l2p.find("?")]
				self.actionDirect (PA(dev.id, {"setting":setting}), "ZP_Pandora")
			elif uri_siriusxm in l2p:
				#setting = l2p[l2p.find("%3a")+3:l2p.find("?")]
				setting = urllib.unquote(l2p[l2p.find(":")+1:l2p.find("?")])
				self.actionDirect (PA(dev.id, {"setting":setting}), "ZP_SiriusXM")
			elif uri_container in l2p or uri_jffs in l2p or uri_playlist in l2p or uri_file in l2p:
				self.actionDirect (PA(dev.id, {"setting":l2p, "mode":mode}), "ZP_Container")
			elif uri_sonos_radio in l2p:
				self.actionDirect (PA(dev.id, {"setting":l2p}), "ZP_SonosRadio")
			else:
				indigo.server.log ("I do not know what to do with Favorite: %s" % l2p)

		elif action == "ZP_DumpURI":
			MediaInfo = self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "GetMediaInfo", "")
			PositionInfo = self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "GetPositionInfo", "")
			self.logSep(0)
			indigo.server.log("ZonePlayer: %s, %s" % (zoneIP, dev.name))
			indigo.server.log("MediaInfo: %s" % MediaInfo)
			indigo.server.log("PositionInfo: %s" % PositionInfo)
			self.logSep(0)
	
	def actionPandoraThumbs(self, pluginAction, action):
		dev = indigo.devices[pluginAction.deviceId]
		zoneIP = dev.pluginProps["address"]

		if dev.states["GROUP_Coordinator"] == "false":
			Coordinator = dev.states["GROUP_Name"]
			for idev in indigo.devices.iter("self.ZonePlayer"):
				if idev.states["GROUP_Coordinator"] == "true" and idev.states["GROUP_Name"] == Coordinator:
					zoneIP = idev.pluginProps["address"]
					dev = indigo.devices[idev.id]
					break

		if uri_pandora in dev.states["ZP_CurrentURI"]:
			(x, y) = dev.states["ZP_CurrentURI"].split(':')
			(stationId, z) = y.split('?')
			for item in Sonos_Pandora:
				if item[0] == stationId:
					PandoraStation = item[1]
					PandoraEmailAddress = item[2]
					break
			trackToken = self.parsePandoraToken(dev.states["ZP_CurrentTrackURI"])

			if PandoraEmailAddress == self.PandoraEmailAddress:
				PandoraPassword = self.PandoraPassword
			elif PandoraEmailAddress == self.PandoraEmailAddress2:
				PandoraPassword = self.PandoraPassword2

			pandora = Pandora()
			pandora.authenticate(PandoraEmailAddress, PandoraPassword)

			if action == "thumbs_up":
				thumbAction = "Thumbs Up"
				feedback = True
			elif action == "thumbs_down":
				thumbAction = "Thumbs Down"
				feedback = False

			try:
				thumb_status = pandora.add_feedback(stationId, trackToken, feedback)
				partist = thumb_status['artistName']
				ptrack = thumb_status['songName']
				if action == "thumbs_down":
					self.actionDirect (PA(dev.id), "Next")

				indigo.server.log("[%s] %s for station: %s, artist: %s, track: %s on ZonePlayer: %s" % (time.asctime(), thumbAction, PandoraStation, partist, ptrack, dev.name))
			except:
				self.plugin.errorLog("[%s] Unable to %s track on ZonePlzyer: %s" % (time.asctime(), thumbAction, dev.name))

		else:
			self.plugin.errorLog("[%s] Pandora not actively playing on ZonePlayer: %s" % (time.asctime(), dev.name))
	

	def actionStates(self, pluginAction, action):
		global SavedState
		if action == "saveStates":
			SavedState = []
			for dev in indigo.devices.iter("self.ZonePlayer"):
			  if dev.enabled == True and dev.pluginProps["model"] != SONOS_SUB:
				#ZP  = self.parseBrowseNumberReturned(self.SOAPSend (dev.pluginProps["address"], "/MediaServer", "/ContentDirectory", "Browse", "<ObjectID>Q:0</ObjectID><BrowseFlag>BrowseDirectChildren</BrowseFlag><Filter></Filter><StartingIndex>0</StartingIndex><RequestedCount>1000</RequestedCount><SortCriteria></SortCriteria>"))
				ZP = ""
				ZP_CurrentURIMetaData = self.parseDirty(self.SOAPSend (dev.pluginProps["address"], "/MediaRenderer", "/AVTransport", "GetMediaInfo", ""), "<CurrentURIMetaData>", "</CurrentURIMetaData>")
				rel_time = self.parseRelTime(dev, self.SOAPSend (dev.pluginProps["address"], "/MediaRenderer", "/AVTransport", "GetPositionInfo", ""))
				SavedState.append((dev.states['ZP_LocalUID'], dev.states['Q_Crossfade'], dev.states['Q_Repeat'], dev.states['Q_Shuffle'], dev.states['ZP_MUTE'], dev.states['ZP_STATE'], dev.states['ZP_VOLUME'], dev.states['ZP_CurrentURI'], ZP_CurrentURIMetaData, dev.states['ZP_CurrentTrack'], dev.states['GROUP_Coordinator'], ZP, rel_time, dev.states['ZonePlayerUUIDsInGroup']))
		elif action == "restoreStates":
			pass

	def actionAnnouncement(self, pluginAction, action):
		global SavedState
		global actionBusy

		actionBusy = 1

		self.actionStates(pluginAction, "saveStates")

		zp_volume = self.plugin.substitute(pluginAction.props.get("zp_volume"))

		# Preserve existing group structure if Group Coordinator Only is selected in action

		try:
			gc_only = pluginAction.props.get("gc_only")
		except:
			gc_only = False

		# need this until group announcement actions are merged
		if action == "announcement":
			gc_only = False

		AnnouncementZones = []
		if gc_only == False:
			x = 1
			while x <= 12:
				ivar = 'zp' + str(x)
				if pluginAction.props.get(ivar) not in ["", None, "00000"]: 
					AnnouncementZones.append(pluginAction.props.get(ivar))
				x = x + 1
		else:
			dev = indigo.devices[int(pluginAction.props.get("zp1"))]
			if dev.states["GROUP_Coordinator"] == "true":
				AnnouncementZones.append(dev.id)
			else:
				# if selected ZonePlayer is not master of a group, find the master
				Coordinator = dev.states["GROUP_Name"]
				for idev in indigo.devices.iter("self.ZonePlayer"):
					if idev.states["GROUP_Coordinator"] == "true" and idev.states["GROUP_Name"] == Coordinator:
						AnnouncementZones.append(idev.id)
						break

		if action == "announcement":
			announcement = self.plugin.substitute(pluginAction.props.get("setting"), validateOnly=False)
			zp_input = pluginAction.props.get("zp_input")

			indigo.server.log("Announcement: %s, Volume: %s, Line-In: %s" % (announcement, zp_volume, zp_input))

			dev_src = indigo.devices[int(zp_input)]
			dev_src_LocalUID = dev_src.states['ZP_LocalUID']

			if dev_src.states['ZP_AIName'] == "":
				indigo.server.log ("No Line-In Available...")
				actionBusy = 0
				return

			for item in AnnouncementZones:
				dev = indigo.devices[int(item)]
				zoneIP = dev.pluginProps["address"]

				# if member of group, remove from group
				# no action necessary since uri will change to line-in
				if dev.states['GROUP_Coordinator'] == "false":
					self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "BecomeCoordinatorOfStandaloneGroup", "")
				# change input to line-in
				self.plugin.debugLog(u"Playing LineIn: %s..." % dev_src.states["ZP_AIName"])
				self.plugin.debugLog("Sonos Action: Line-In")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon-stream:"+str(dev_src_LocalUID)+"</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
				# set announcment volume
				self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetVolume", "<Channel>Master</Channel><DesiredVolume>"+str(zp_volume)+"</DesiredVolume>")
				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Play", "<Speed>1</Speed>")
				# un-mute ZonePlayer
				self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>0</DesiredMute>")

			# make announcement
			self.plugin.debugLog("Making Announcement...")
			self.plugin.sleep(0.5)
			indigo.server.speak(" ", waitUntilDone=True)
			indigo.server.speak(announcement, waitUntilDone=True)
			self.plugin.sleep(1.0)

		elif action == "announcementMP3":
			if pluginAction.props.get("ttsORfile") == "TTS":
				announcement = self.plugin.substitute(pluginAction.props.get("setting"), validateOnly=False)
				zp_language = pluginAction.props.get("language")
				tts = gTTS(text=announcement, lang=zp_language)
				tts.save('announcement.mp3')
				s_announcement = "announcement.mp3"
				tts_delay = 0

			elif pluginAction.props.get("ttsORfile") == "IVONA":
				announcement = self.plugin.substitute(pluginAction.props.get("IVONA_setting"), validateOnly=False)

				v = pyvona.pyvona.create_voice(self.IVONAaccessKey, self.IVONAsecretKey)
				v.codec = 'mp3'
				v.voice_name = IVONAVoices[int(pluginAction.props.get("IVONA_voice"))][1]
				v.sentence_break = int(pluginAction.props.get("IVONA_sentence_break"))
				v.speech_rate = pluginAction.props.get("IVONA_speech_rate")
				v.fetch_voice(announcement, 'announcement')
				s_announcement = "announcement.mp3"
				tts_delay = 0.5
				# small delay to allow MP3 file creation completion
				self.plugin.sleep(0.5)

			elif pluginAction.props.get("ttsORfile") == "POLLY":
				announcement = self.plugin.substitute(pluginAction.props.get("POLLY_setting"), validateOnly=False)
				client = boto3.client('polly', aws_access_key_id=self.PollyaccessKey, aws_secret_access_key=self.PollysecretKey, region_name='us-east-1')
				response = client.synthesize_speech(OutputFormat='mp3', Text=announcement, VoiceId=pluginAction.props.get("POLLY_voice"))
				if "AudioStream" in response:
					with closing(response["AudioStream"]) as stream:
						data = stream.read()
						f = open("announcement.mp3", "w+")
						f.write(data)
						f.close()
				s_announcement = "announcement.mp3"
				tts_delay = 0.5

			elif pluginAction.props.get("ttsORfile") == "APPLE":
				announcement = self.plugin.substitute(pluginAction.props.get("APPLE_setting"), validateOnly=False)
				sp = NSSpeechSynthesizer.alloc().initWithVoice_(pluginAction.props.get("APPLE_voice"))
				ru = NSURL.fileURLWithPath_("./announcement.aiff")
				sp.startSpeakingString_toURL_(announcement, ru)
				s_announcement = "announcement.aiff"
				tts_delay = 0.5
				# small delay to allow MP3 file creation completion
				self.plugin.sleep(0.5)

			elif pluginAction.props.get("ttsORfile") == "MICROSOFT":
				announcement = self.plugin.substitute(pluginAction.props.get("MICROSOFT_setting"), validateOnly=False)
				language = pluginAction.props.get("MICROSOFT_voice")
				statinfo = self.MicrosoftTranslate(announcement, language)
				s_announcement = "announcement.mp3"
				tts_delay = 0.5
				if statinfo == False:
					self.plugin.errorLog ("Microsoft Translate Error")
					return

			else:
				announcement = "FILE [" + pluginAction.props.get("sound_file") + "]"
				os.system("cp -pr \"" + self.SoundFilePath + "/" + pluginAction.props.get("sound_file") + "\" announcement.mp3")
				s_announcement = "announcement.mp3"
				tts_delay = 0


			indigo.server.log("Announcement: %s, Volume: %s" % (announcement, zp_volume))

			GM = indigo.devices[int(AnnouncementZones[0])]
			zoneIP = GM.pluginProps["address"]

			if gc_only == True:
				group_volume = self.parseCurrentVolume (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupVolume", ""))
				group_mute = self.parseCurrentMute (self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "GetGroupMute", ""))

			if gc_only == False:
				# set standalone
				self.plugin.debugLog ("Announcement: set standalone")
				for item in AnnouncementZones:
					dev = indigo.devices[int(item)]
					self.actionDirect (PA(dev.id), "setStandalone")
				# add announcement zones to group
				self.plugin.debugLog ("Announcement: add announcement zones to group")
				itemcount = 0	
				for item in AnnouncementZones:
					dev = indigo.devices[int(item)]
					if itemcount > 0:
						self.actionDirect (PA(GM.id, {'setting':dev.id}), "addPlayerToZone")
					itemcount = itemcount + 1
			else:
				self.actionDirect(PA(GM.id), "Stop")

			# set volume
			self.plugin.debugLog ("Announcement: set volume")
			if gc_only == False:
				for item in AnnouncementZones:
					dev = indigo.devices[int(item)]
					self.actionDirect (PA(dev.id, {'setting':zp_volume}), "Volume")
			else:
				self.actionDirect (PA(GM.id, {'setting':zp_volume}), "GroupVolume")
				self.actionDirect (PA(GM.id), "GroupMuteOff")

			count = 0
			success = 0
			while count < 5 and success == 0:
				try:
					if "mp3" in  s_announcement:
						audio = MP3("./" + s_announcement)
					elif "aiff" in s_announcement:
						audio = AIFF("./" + s_announcement)
					success = 1
				except:
					self.plugin.sleep(0.5)
					count = count + 1

			if success == 1:
				indigo.server.log("Announcement Length: %s" % audio.info.length)

				self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>http://" + self.HTTPServer + ":" + self.HTTPStreamingPort + "/" + s_announcement + "</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")

				# turn off queue repeat
				self.actionDirect (PA(GM.id, {'setting':0}), "Q_Repeat")
			
				self.plugin.sleep(1)
			
				self.actionDirect (PA(GM.id), "Play")
				self.plugin.sleep(tts_delay + audio.info.length)
			else:
				self.plugin.errorLog("Unable to read MP3 file.  Announcement aborted.")

		# restore state

		updatedURI = {}
		for item in AnnouncementZones:
			dev = indigo.devices[int(item)]		
			if gc_only == False:
				self.actionDirect (PA(dev.id), "setStandalone")
			updatedURI[dev.id] = 0
			self.plugin.sleep(.5)
		
		for item in AnnouncementZones:
			dev = indigo.devices[int(item)]
			zoneIP = dev.pluginProps["address"]

			for state in SavedState:
				if state[0] == dev.states['ZP_LocalUID']:
					self.plugin.debugLog("Restore States: %s:%s:%s" % (state[0], dev.states['ZP_LocalUID'], dev.name))
					self.plugin.debugLog("01:Q_Crossfade:            %s" % state[1])
					self.plugin.debugLog("02:Q_Repeat:               %s" % state[2])
					self.plugin.debugLog("03:Q_Shuffle:              %s" % state[3])
					self.plugin.debugLog("04:ZP_MUTE:                %s" % state[4])
					self.plugin.debugLog("05:ZP_STATE:               %s" % state[5])
					self.plugin.debugLog("06:ZP_VOLUME:              %s" % state[6])
					self.plugin.debugLog("07:ZP_CurrentURI:          %s" % state[7])
					self.plugin.debugLog("08:CurrentURI MetaData:    %s" % state[8])
					self.plugin.debugLog("09:ZP_CurrentTrack:        %s" % state[9])
					self.plugin.debugLog("10:GROUP_Coordinator:      %s" % state[10])
					self.plugin.debugLog("12:Relative Time:          %s" % state[12])
					self.plugin.debugLog("13:ZonePlayerUUIDsInGroup: %s" % state[13])

					if gc_only == False:
						# restore group membership
						if state[10] == "true":
							self.plugin.debugLog("%s: Restore URI: %s" % (state[0], state[7]))

							# Restore Queue from Saved Playlist
							if int(dev.states['Q_Number']) == 0:
								self.actionDirect (PA(dev.id), "Q_Clear")
							else:
								for plist in Sonos_Playlists:
									if dev.states['Q_ObjectID'] == plist[2]:
										l2p = plist[0]
								self.actionDirect (PA(dev.id), "Q_Clear")
								self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "AddURIToQueue", "<EnqueuedURI>"+l2p+"</EnqueuedURI><EnqueuedURIMetaData></EnqueuedURIMetaData><DesiredFirstTrackNumberEnqueued>0</DesiredFirstTrackNumberEnqueued><EnqueueAsNext>1</EnqueueAsNext>")

							SendVar = "<CurrentURI>"+state[7].replace("&", "&amp;")+"</CurrentURI><CurrentURIMetaData>" + state[8].decode('utf-8') + "</CurrentURIMetaData>" 
							self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", SendVar.encode('utf-8'))
							updatedURI[dev.id] = 1

							ZonePlayerUUIDsInGroup = state[13].split(',')
							for ZonePlayerUUID in ZonePlayerUUIDsInGroup:
								if state[0] != ZonePlayerUUID:
									for rdev in indigo.devices.iter("self.ZonePlayer"):
										if ZonePlayerUUID == rdev.states['ZP_LocalUID']:
											indigo.server.log("add zone to group: %s->%s" % (rdev.name, dev.name))
											self.SOAPSend (rdev.pluginProps['address'], "/MediaRenderer", "/AVTransport", "SetAVTransportURI", "<CurrentURI>x-rincon:"+str(state[0])+"</CurrentURI><CurrentURIMetaData></CurrentURIMetaData>")
											updatedURI[rdev.id] = 1
											self.plugin.sleep(.5)

						# restore mute
						self.plugin.debugLog("%s: Restore Mute: %s" % (state[0], state[4]))
						if (state[4] == "0"):
							self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>0</DesiredMute>")
						else:
							self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetMute", "<Channel>Master</Channel><DesiredMute>1</DesiredMute>")
						# restore volume
						self.plugin.debugLog("%s: Restore Volume: %s" % (state[0], state[6]))
						self.SOAPSend (zoneIP, "/MediaRenderer", "/RenderingControl", "SetVolume", "<Channel>Master</Channel><DesiredVolume>"+state[6]+"</DesiredVolume>")
					else:
						self.actionDirect (PA(dev.id, {'setting':group_volume}), "GroupVolume")
						self.SOAPSend (zoneIP, "/MediaRenderer", "/GroupRenderingControl", "SetGroupMute", "<DesiredMute>" + group_mute + "</DesiredMute>")

					# resture uri
					if updatedURI[dev.id] == 0:
						self.plugin.debugLog("%s: Restore URI: %s" % (state[0], state[7]))
						SendVar = "<CurrentURI>"+state[7].replace("&", "&amp;")+"</CurrentURI><CurrentURIMetaData>" + state[8].decode('utf-8') + "</CurrentURIMetaData>" 
						self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "SetAVTransportURI", SendVar.encode('utf-8'))
						updatedURI[dev.id] = 1
					
		self.plugin.sleep(1)

		for item in AnnouncementZones:
			dev = indigo.devices[int(item)]
			zoneIP = dev.pluginProps["address"]
			CurrentTransportActions = self.parseCurrentTransportActions (self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "GetCurrentTransportActions", ""))

			for state in SavedState:
				if state[0] == dev.states['ZP_LocalUID']:
					# ZonePlayer is Group Coordinator and has a Current URI
					if state[10] == "true":
						if uri_music in dev.states['ZP_CurrentURI'] and int(dev.states['Q_Number']) > 0:
							QueueReady = True
						else:
							QueueReady = False

						if CurrentTransportActions.find("Seek") >= 0 and QueueReady == True:
							# restore track number
							self.plugin.debugLog("%s: Restore Track Number: %s" % (state[0], state[9]))
							self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Seek", "<Unit>TRACK_NR</Unit><Target>"+state[9]+"</Target>")
							# restore rel time
							self.plugin.debugLog("%s: Restore Relative Time: %s" % (state[0], state[12]))
							self.SOAPSend (zoneIP, "/MediaRenderer", "/AVTransport", "Seek", "<Unit>REL_TIME</Unit><Target>"+state[12]+"</Target>")
						
							# restore Q_Repeat
							if state[2] == "off":
								Q_setting = 0
							else:
								Q_setting = 1
							self.actionDirect (PA(dev.id, {'setting':Q_setting}), "Q_Repeat")

						# restore play state
						self.plugin.debugLog("%s: Restore Play State: %s" % (state[0], state[5]))
						if (state[5] == "PLAYING"):
							self.actionDirect (PA(dev.id), "Play")
						elif (state[5] == "PAUSED"):
							self.actionDirect (PA(dev.id), "Play")
							self.actionDirect (PA(dev.id), "Pause")
						#else:
						#	self.actionDirect (PA(dev.id), "Stop")

		AnnouncementZones = []
		actionBusy = 0	

	def MicrosoftTranslateAuth(self):
		authUrl = 'https://datamarket.accesscontrol.windows.net/v2/OAuth2-13/'
		scopeUrl = 'http://api.microsofttranslator.com'
		grantType = 'client_credentials'

		postdata = {'grant_type':grantType, 'scope':scopeUrl, 'client_id':self.MSTranslateClientID, 'client_secret':self.MSTranslateClientSecret}
		response = requests.post(authUrl, data=postdata)

		if response.status_code == 200:
			content = json.loads (response.content)
			return (content['access_token'])
		else:
			self.plugin.errorLog("[%s] Cannot authenticate to Microsoft Translate" % time.asctime())
			return (False)

	def MicrosoftTranslateLanguages(self):
		accessToken = self.MicrosoftTranslateAuth()
		if accessToken == False:
			return (False)

		scopeUrl = 'http://api.microsofttranslator.com'
		headers = {'Content-Type':'text/xml', 'Authorization':'Bearer ' + accessToken}
		url = scopeUrl + '/V2/Http.svc/GetLanguagesForSpeak'
		response = requests.get(url, headers=headers)

		langCodes = []
		Languages = ET.fromstring(response.content)
		for lang in Languages:
			langCodes.append(lang.text)
		languageCodes = str(langCodes).replace("'",'"')

		#self.myLocale = self.getLocale()
		#if self.myLocale == None:
		self.myLocale = 'en'

		url = scopeUrl + '/V2/Ajax.svc/GetLanguageNames?locale=' + self.myLocale + '&languageCodes=' + languageCodes
		response = requests.post(url, headers=headers)

		name_code = dict(zip(langCodes, eval(response.content)))
		indigo.server.log("Loaded Microsoft Translate Voices... [%s]" % len(name_code))

		return (name_code)

	def MicrosoftTranslate(self, announcement, language):
		authUrl = 'https://datamarket.accesscontrol.windows.net/v2/OAuth2-13/'
		scopeUrl = 'http://api.microsofttranslator.com'
		speakUrl = 'http://api.microsofttranslator.com/V2/Http.svc/Speak'
		grantType = 'client_credentials'

		accessToken = self.MicrosoftTranslateAuth()
		if accessToken == False:
			return (False)

		headers = {'Content-Type':'audio/mp3', 'Authorization':'Bearer ' + accessToken}
		url = speakUrl + '?text=' + announcement + '&language=' + language + '&format=audio/mp3&options=MaxQuality'

		with open ('announcement.mp3', 'wb') as handle:
			response = requests.get(url, headers=headers, stream=True)

			if response.ok:
				for block in response.iter_content(1024):
					handle.write(block)
				return (True)
			else:
				return (False)

	def getReferencePlayerIP(self):
		return soco.discover().pop().ip_address

	######################################################################################
	# Plugin Preferences
	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		if not userCancelled:		
			self.plugin.debugLog(u"[%s] Getting plugin preferences." % time.asctime())

			try:
				self.plugin.debug = self.plugin.pluginPrefs[u'showDebugInLog']
			except:
				self.plugin.debug = False

			try:
				self.plugin.xmlDebug = self.plugin.pluginPrefs[u'showXMLInLog']
			except:
				self.plugin.xmlDebug = False

			try:
				self.plugin.eventsDebug = self.plugin.pluginPrefs[u'showEventsInLog']
			except:
				self.plugin.eventsDebug = False

			try:
				self.plugin.stateUpdatesDebug = self.plugin.pluginPrefs[u'showStateUpdatesInLog']
			except:
				self.plugin.stateUpdatesDebug = False

			if self.rootZPIP != self.plugin.pluginPrefs["rootZPIP"]:
				self.rootZPIP = self.plugin.pluginPrefs["rootZPIP"]
				if self.rootZPIP == 'auto':
					self.rootZPIP = self.getReferencePlayerIP()
					indigo.server.log("Using Reference ZonePlayer IP: %s" % self.rootZPIP)
				if self.rootZPIP != None:
					self.getSonosFavorites()
					self.getPlaylistsDirect()
					self.getRT_FavStationsDirect()
					# Retrieve Sonos Device ID for Music API
					url = u"http://" + self.rootZPIP + ":1400/status/zp"
					response = requests.get(url)
					if response.ok:
						root = ET.fromstring(response.content)
						self.SonosDeviceID = root.findtext('.//SerialNumber')
					else:
						self.plugin.errorLog("[%s] Cannot retrieve SerialNumber from Root ZonePlayer: %s" % (time.asctime(), self.rootZPIP))
				else:
					self.plugin.errorLog("[%s] Reference ZonePlayer IP address invalid." % time.asctime())

			try: self.EventProcessor = self.plugin.pluginPrefs["EventProcessor"]
			except: self.EventProcessor = "SoCo"
			
			try: self.EventIP = self.plugin.pluginPrefs["EventIP"]
			except: self.plugin.errorLog("[%s] Could not retrieve Event Listener IP address." % time.asctime())

			try:
				if (self.HTTPStreamingIP != self.plugin.pluginPrefs["HTTPStreamingIP"]) or (self.HTTPStreamingPort != self.plugin.pluginPrefs["HTTPStreamingPort"]):
					self.HTTPStreamingIP = self.plugin.pluginPrefs["HTTPStreamingIP"]
					self.HTTPStreamingPort = self.plugin.pluginPrefs["HTTPStreamingPort"]

					self.HTTPSTreamerOn = False
					v = Thread(target=self.HTTPStreamer)
					v.setDaemon(True)
					v.start()
			except:
				self.plugin.errorLog("[%s] HTTPStreamer not functioning." % time.asctime())

			try:
				if self.SoundFilePath != self.plugin.pluginPrefs["SoundFilePath"]:
					self.SoundFilePath = self.plugin.pluginPrefs["SoundFilePath"]
					if self.SoundFilePath != None and self.SoundFilePath != "":
						self.getSoundFiles()
			except:
				self.plugin.errorLog("[%s] Could not retrieve SoundFilePath." % time.asctime())

			try:
				if (self.Pandora != self.plugin.pluginPrefs[u'Pandora']) or \
				(self.PandoraEmailAddress != self.plugin.pluginPrefs[u'PandoraEmailAddress']) or \
				(self.PandoraPassword != self.plugin.pluginPrefs[u'PandoraPassword']) or \
				(self.PandoraNickname != self.plugin.pluginPrefs[u'PandoraNickname']):
					self.Pandora = self.plugin.pluginPrefs[u'Pandora']
					self.PandoraEmailAddress = self.plugin.pluginPrefs[u'PandoraEmailAddress']
					self.PandoraPassword = self.plugin.pluginPrefs[u'PandoraPassword']
					self.PandoraNickname = self.plugin.pluginPrefs[u'PandoraNickname']
					if self.Pandora:
						self.getPandora(self.PandoraEmailAddress, self.PandoraPassword, self.PandoraNickname)
			except:
				self.plugin.errorLog("[%s] Could not retrieve Pandora credentials." % time.asctime())

			try:
				if (self.Pandora2 != self.plugin.pluginPrefs[u'Pandora2']) or \
				(self.PandoraEmailAddress2 != self.plugin.pluginPrefs[u'PandoraEmailAddress2']) or \
				(self.PandoraPassword2 != self.plugin.pluginPrefs[u'PandoraPassword2']) or \
				(self.PandoraNickname2 != self.plugin.pluginPrefs[u'PandoraNickname2']):
					self.Pandora2 = self.plugin.pluginPrefs[u'Pandora2']
					self.PandoraEmailAddress2 = self.plugin.pluginPrefs[u'PandoraEmailAddress2']
					self.PandoraPassword2 = self.plugin.pluginPrefs[u'PandoraPassword2']
					self.PandoraNickname2 = self.plugin.pluginPrefs[u'PandoraNickname2']
					if self.Pandora2:
						self.getPandora(self.PandoraEmailAddress2, self.PandoraPassword2, self.PandoraNickname2)
			except:
				self.plugin.errorLog("[%s] Could not retrieve secondary Pandora credentials." % time.asctime())

			try:
				if (self.SiriusXM != self.plugin.pluginPrefs[u'SiriusXM']) or \
				(self.SiriusXMID != self.plugin.pluginPrefs[u'SiriusXMID']) or \
				(self.SiriusXMPassword != self.plugin.pluginPrefs[u'SiriusXMPassword']):
					self.SiriusXM = self.plugin.pluginPrefs[u'SiriusXM']
					self.SiriusXMID = self.plugin.pluginPrefs[u'SiriusXMID']
					self.SiriusXMPassword = self.plugin.pluginPrefs[u'SiriusXMPassword']
					if self.SiriusXM:
						self.getSiriusXM()
			except:
				self.plugin.errorLog("[%s] Could not retrieve SiriusXM parameters." % time.asctime())

			try:
				if (self.IVONA != self.plugin.pluginPrefs[u'IVONA']) or \
				(self.IVONAaccessKey != self.plugin.pluginPrefs[u'IVONAaccessKey']) or \
				(self.IVONAsecretKey != self.plugin.pluginPrefs[u'IVONAsecretKey']):
					self.IVONA = self.plugin.pluginPrefs[u'IVONA']
					if self.IVONA:
						self.IVONAaccessKey = self.plugin.pluginPrefs[u'IVONAaccessKey']
						self.IVONAsecretKey = self.plugin.pluginPrefs[u'IVONAsecretKey']
					if self.IVONA:
						self.IVONAVoices()
			except:
				self.plugin.errorLog("[%s] Could not retrieve IVONA parameters." % time.asctime())

			try:
				if (self.Polly != self.plugin.pluginPrefs[u'Polly']) or \
				(self.PollyaccessKey != self.plugin.pluginPrefs[u'PollyaccessKey']) or \
				(self.PollysecretKey != self.plugin.pluginPrefs[u'PollysecretKey']):
					self.Polly = self.plugin.pluginPrefs[u'Polly']
					if self.Polly:
						self.PollyaccessKey = self.plugin.pluginPrefs[u'PollyaccessKey']
						self.PollysecretKey = self.plugin.pluginPrefs[u'PollysecretKey']
					if self.Polly:
						self.PollyVoices()
			except:
				self.plugin.errorLog("[%s] Could not retrieve Polly parameters." % time.asctime())

			try:
				if (self.MSTranslate != self.plugin.pluginPrefs[u'MSTranslate']) or \
				(self.MSTranslateClientID != self.plugin.pluginPrefs[u'MSTranslateClientID']) or \
				(self.MSTranslateClientSecret != self.plugin.pluginPrefs[u'MSTranslateClientSecret']):
					self.MSTranslate = self.plugin.pluginPrefs[u'MSTranslate']
					if self.MSTranslate == True:
						self.MSTranslateClientID = self.plugin.pluginPrefs[u'MSTranslateClientID']
						self.MSTranslateClientSecret = self.plugin.pluginPrefs[u'MSTranslateClientSecret']
					if self.MSTranslate == True:
						self.MSTranslateVoices = self.MicrosoftTranslateLanguages()
			except:
				self.plugin.errorLog("[%s] Could not retrieve MSTranslate parameters." % time.asctime())

			indigo.server.log("[%s] Processed plugin preferences." % time.asctime())
			return True


	######################################################################################
	# UI Validation
	def validatePrefsConfigUi(self, valuesDict):

		self.plugin.debugLog(u"Vaidating Plugin Configuration")
		errorsDict = indigo.Dict()
		if valuesDict[u"rootZPIP"] == "":
			errorsDict[u"rootZPIP"] = u"Please enter a reference ZonePlayer IP Address."
		if valuesDict[u"EventProcessor"] == "":
			errorsDict[u"EventProcessor"] = u"Please select an Event Processsor."
		if valuesDict[u"EventIP"] == "":
			errorsDict[u"EventIP"] = u"Please select an Event Listener IP Address."
		if valuesDict[u"HTTPStreamingIP"] == "":
			errorsDict[u"rootZPIP"] = u"Please enter an HTTP Streaming IP Address."
		if valuesDict[u"HTTPStreamingPort"] == "":
			errorsDict[u"rootZPIP"] = u"Please enter an HTTP Streaming Port."
		if valuesDict[u"Pandora"] == "True":
			if valuesDict[u"PandoraEmailAddress"] == "":
				errorsDict[u"PandoraEmailAddress"] = u"Please enter a Pandora Email Address."
			if valuesDict[u"PandoraPassword"] == "":
				errorsDict[u"PandoraPassword"] = u"Please enter a Pandora Password."
		if valuesDict[u"SiriusXM"] == "True":
			if valuesDict[u"SiriusXMID"] == "":
				errorsDict[u"SiriusXMID"] = u"Please enter a SiriusXM ID."
			if valuesDict[u"SiriusXMPassword"] == "":
				errorsDict[u"SiriusXMPassword"] = u"Please enter a SiriusXM Password."
		if valuesDict[u"IVONA"] == "True":
			if valuesDict[u"IVONAaccessKey"] == "":
				errorsDict[u"IVONAaccessKey"] = u"Please enter an IVONA Access Key."
			if valuesDict[u"IVONAsecretKey"] == "":
				errorsDict[u"IVONAsecretKey"] = u"Please enter an IVONA Secret Key."
		if valuesDict[u"Polly"] == "True":
			if valuesDict[u"PollyaccessKey"] == "":
				errorsDict[u"PollyaccessKey"] = u"Please enter an Polly Access Key."
			if valuesDict[u"PollysecretKey"] == "":
				errorsDict[u"PollysecretKey"] = u"Please enter an Polly Secret Key."
		if valuesDict[u"MSTranslate"] == "True":
			if valuesDict[u"MSTranslateClientID"] == "":
				errorsDict[u"MSTranslateClientID"] = u"Please enter a Microsoft Translate Client ID."
			if valuesDict[u"MSTranslateClientSeret"] == "":
				errorsDict[u"MSTranslateClientSecret"] = u"Please enter an Microsoft Translate Client Secret."

		if len(errorsDict) > 0:
			self.plugin.errorLog(u"\t Validation Errors")
			return (False, valuesDict, errorsDict)
		else:			
			self.plugin.debugLog(u"\t Validation Succesful")
			return (True, valuesDict)


	######################################################################################

	def getZPDeviceList(self, filter=""):
		array = []
		if filter == "withNone":
			array.append(("00000", "No Selection"))
		for dev in indigo.devices.iter("self.ZonePlayer"):
			array.append((dev.id, dev.states['ZP_ZoneName']))
		return array
	
	def getZP_LIST(self, filter=""):
		array = []
		for plist in Sonos_Playlists:
			array.append((plist[0], plist[1]))
		return array

	def getZP_LIST_PlaylistObjects(self, filter=""):
		array = []
		for plist in Sonos_Playlists:
			array.append((plist[2], plist[1]))
		return array	

	def getZP_LineIn(self, filter=""):
		return Sonos_LineIn

	def getZP_SonosFavorites(self, filter=""):
		array=[]
		for title in Sonos_Favorites:
			array.append((title[4], title[1]))
		return array

	def getZP_RT_FavStations(self, filter=""):
		return Sonos_RT_FavStations

	def getZP_Pandora(self, filter=""):
		array = []
		for station in Sonos_Pandora:
			array.append((station[0], station[3] + ": " + station[1]))
		array.sort(key=lambda x:x[1])
		return array

	def getZP_SiriusXM(self, filter=""):
		array = []
		for channel in Sonos_SiriusXM:
			array.append((channel[1], channel[2]))
		#	array.append((channel[3], str(channel[0]).zfill(3) + " | " + channel[4]))
		#array.sort(key=lambda x:x[1])
		return array

	def getZP_SoundFiles(self, filter=""):
		return Sound_Files

	def getIVONAVoices(self, filter=""):
		array = []
		for voice in IVONAVoices:
			array.append((voice[0], voice[4] + " | " + voice[1]))
		array.sort(key=lambda x:x[1])
		return array

	def getPollyVoices(self, filter=""):
		array = []
		for voice in PollyVoices:
			array.append((voice[0], voice[4] + " | " + voice[1]))
		array.sort(key=lambda x:x[1])
		return array

	def getAppleVoices(self, filter=""):
		array = []
		for voice in NSVoices:
			name = NSSpeechSynthesizer.attributesForVoice_(voice)['VoiceName']
			locale = re.split('-|_', NSSpeechSynthesizer.attributesForVoice_(voice)['VoiceLocaleIdentifier'])
			try:
				vl = language_codes.languages[locale[0]].encode('utf-8')
			except:
				vl = locale[0]
			try:
				vc = language_codes.countries[locale[1]].encode('utf-8')
			except:
				vc = locale[1]
			array.append((voice,  vc + ', ' +  vl + ' | ' + name))
		array.sort(key=lambda x:x[1])
		return array

	def getMicrosoftLanguages(self, filter=""):
		array = []
		for code in self.MSTranslateVoices:
			array.append((code, self.MSTranslateVoices[code]))
		array.sort(key=lambda x:x[1])
		return array

	######################################################################################
	# Utiliies

	def cleanString(self, in_string):
		in_string = in_string.replace("&", "&amp;amp;")
		in_string = in_string.replace("'", "&apos;")
		return in_string

	def restoreString(self, in_string, filter):
		in_string = in_string.replace("&amp;apos;", "\'")
		if filter == 0:
			in_string = in_string.replace("&amp;amp;", "&")
			in_string = in_string.replace("&amp;", "&")
		in_string = in_string.replace("&quot;", "\"")
		in_string = in_string.replace("&lt;", "<")
		in_string = in_string.replace("&gt;", ">")
		in_string = in_string.replace("&apos;", "'")
		return in_string

	def logSep(self, debug):
		if debug:
			self.plugin.debugLog(u"---------------------------------------------")
		else:
			indigo.server.log(u"---------------------------------------------")

	def updateStateOnServer(self, dev, state, value):
		if self.plugin.stateUpdatesDebug:
			self.plugin.debugLog(u"\t Updating Device: %s, State: %s, Value: %s" % (dev.name, state, value))
		GROUP_Coordinator = dev.states['GROUP_Coordinator']
		if GROUP_Coordinator == "false" and state in ZoneGroupStates:
			pass
		else:
			if value == None or value == "None":
				dev.updateStateOnServer(state, "")
			else:
				dev.updateStateOnServer(state, value.encode('utf-8'))

		# Replicate states to slave ZonePlayers
		if state in ZoneGroupStates and dev.states['GROUP_Coordinator'] == "true" and dev.states['ZonePlayerUUIDsInGroup'].find(",") != -1:
			self.plugin.debugLog("Replicate state to slave ZonePlayers...")
			ZonePlayerUUIDsInGroup = dev.states['ZonePlayerUUIDsInGroup'].split(',')
			for rdev in indigo.devices.iter("self.ZonePlayer"):
				SlaveUID = rdev.states['ZP_LocalUID']
				GROUP_Coordinator = rdev.states['GROUP_Coordinator']
				if SlaveUID != dev.states['ZP_LocalUID'] and GROUP_Coordinator == "false" and SlaveUID in ZonePlayerUUIDsInGroup:
					if state == "ZP_CurrentURI":
						value = uri_group + dev.states['ZP_LocalUID']
					if self.plugin.stateUpdatesDebug:
						self.plugin.debugLog(u"\t Updating Device: %s, State: %s, Value: %s" % (rdev.name, state, value))
					if value == None or value == "None":
						rdev.updateStateOnServer(state, "")
					else:
						rdev.updateStateOnServer(state, value.encode('utf-8'))

	def updateStateOnSlaves(self, dev):
			self.plugin.debugLog("Update all states to slave ZonePlayers...")
			ZonePlayerUUIDsInGroup = dev.states['ZonePlayerUUIDsInGroup']
			for rdev in indigo.devices.iter("self.ZonePlayer"):
				SlaveUID = rdev.states['ZP_LocalUID']
				GROUP_Coordinator = rdev.states['GROUP_Coordinator']
				# Do not update if you are yourself, not a slave, and not in the group
				if SlaveUID != dev.states['ZP_LocalUID'] and GROUP_Coordinator == "false" and SlaveUID in ZonePlayerUUIDsInGroup:
					for state in list(ZoneGroupStates):
						if state == "ZP_CurrentURI":
							value = uri_group + dev.states['ZP_LocalUID']
						else:
							value = dev.states[state]
						if self.plugin.stateUpdatesDebug:
							self.plugin.debugLog(u"\t Updating Slave Device: %s, State: %s, Value: %s" % (rdev.name, state, value))
						rdev.updateStateOnServer(state, value)
					rdev.updateStateOnServer("ZP_ART", dev.states['ZP_ART'])
					try:
						shutil.copy2("/Library/Application Support/Perceptive Automation/images/Sonos/"+dev.states['ZP_ZoneName']+"_art.jpg", \
							"/Library/Application Support/Perceptive Automation/images/Sonos/"+rdev.states['ZP_ZoneName']+"_art.jpg")
					except:
						pass

	def copyStateFromMaster(self, dev):
		self.plugin.debugLog("Copy states from master ZonePlayer...")
		(MasterUID,x) = dev.states['GROUP_Name'].split(":")
		for mdev in indigo.devices.iter("self.ZonePlayer"):
			if mdev.states['ZP_LocalUID'] == MasterUID:
				for state in list(ZoneGroupStates):
					if state == "ZP_CurrentURI":
						value = uri_group + mdev.states['ZP_LocalUID']
					else:
						value = mdev.states[state]
					if self.plugin.stateUpdatesDebug:
						self.plugin.debugLog(u"\t Updating Slave Device: %s, State: %s, Value: %s" % (dev.name, state, value))
					dev.updateStateOnServer(state, value)
				dev.updateStateOnServer("ZP_ART", mdev.states['ZP_ART'])
				try:
					shutil.copy2("/Library/Application Support/Perceptive Automation/images/Sonos/"+mdev.states['ZP_ZoneName']+"_art.jpg", \
						"/Library/Application Support/Perceptive Automation/images/Sonos/"+dev.states['ZP_ZoneName']+"_art.jpg")
				except:
					pass
		
	def getLocale(self):
		locale.setlocale(locale.LC_ALL, '')
		return (locale.getdefaultlocale()[0])

	def getSonosFavorites(self):
		global Sonos_Favorites
		list_count = 0
		Sonos_Favorites = []
		ZP  = self.restoreString(self.SOAPSend (self.rootZPIP, "/MediaServer", "/ContentDirectory", "Browse", "<ObjectID>FV:2</ObjectID><BrowseFlag>BrowseDirectChildren</BrowseFlag><Filter></Filter><StartingIndex>0</StartingIndex><RequestedCount>1000</RequestedCount><SortCriteria></SortCriteria>"),1)
		ZPxml = ET.fromstring(ZP)
		iter = ZPxml.getiterator()
		for element in iter:
			if str(element).find("}item") >= 0:
				if element.keys():
					for name, value in element.items():
						if name == "id":
							e_id = value 
				for child in element.getchildren():
					ctag = str(child.tag).split('}')
					if ctag[1] == "title":
						e_title = self.restoreString(child.text,0)
					elif ctag[1] == "res":
						e_res = self.restoreString(child.text,0)
						e_res_clean = child.text
					elif ctag[1] == "resMD":
						e_resMD = child.text
						#indigo.server.log("resMD: %s" % e_resMD)
						#e_class = self.parsePoint(e_resMD, "&lt;upnp:class&gt;", "&lt;/upnp:class&gt;")
					elif ctag[1] == "ordinal":
						e_ordinal = child.text
				Sonos_Favorites.append ((e_res, e_title, e_resMD, e_res_clean, e_id))
				self.plugin.debugLog(u"\tSonos Favorites: %s, %s, %s" % (e_id, e_title, e_res))
				list_count = list_count + 1
		indigo.server.log(u"Loaded Sonos Favorites... [%s]" % list_count)

	def getPlaylistsDirect(self):
		global Sonos_Playlists
		list_count = 0
		Sonos_Playlists = []
		ZP  = self.restoreString(self.SOAPSend (self.rootZPIP, "/MediaServer", "/ContentDirectory", "Browse", "<ObjectID>SQ:</ObjectID><BrowseFlag>BrowseDirectChildren</BrowseFlag><Filter></Filter><StartingIndex>0</StartingIndex><RequestedCount>1000</RequestedCount><SortCriteria></SortCriteria>"),1)
		#self.plugin.debugLog("ZP: %s" % ZP)
		ZPxml = ET.fromstring(ZP)
		iter = ZPxml.getiterator()
		for element in iter:
			if str(element).find("}container") >= 0:
				if element.keys():
					for name, value in element.items():
						if name == "id":
							e_id = value 
				for child in element.getchildren():
					ctag = str(child.tag).split('}')
					if ctag[1] == "title":
						e_title = self.restoreString(child.text,0)
					elif ctag[1] == "res":
						e_res = child.text
				Sonos_Playlists.append ((e_res, e_title, e_id))
				self.plugin.debugLog(u"\tPlaylist: %s, %s, %s" % (e_id, e_title, e_res))
				list_count = list_count + 1
		indigo.server.log(u"Loaded Playlists... [%s]" % list_count)

	def getRT_FavStationsDirect(self):
		global Sonos_RT_FavStations
		list_count = 0
		Sonos_RT_FavStations = []
		ZP  = self.restoreString(self.SOAPSend (self.rootZPIP, "/MediaServer", "/ContentDirectory", "Browse", "<ObjectID>R:0/0</ObjectID><BrowseFlag>BrowseDirectChildren</BrowseFlag><Filter></Filter><StartingIndex>0</StartingIndex><RequestedCount>1000</RequestedCount><SortCriteria></SortCriteria>"),1)
		#self.plugin.debugLog("ZP: %s" % ZP)
		ZPxml = ET.fromstring(ZP)
		iter = ZPxml.getiterator()
		for element in iter:
			if str(element).find("}item") >= 0:
				if element.keys():
					for name, value in element.items():
						if name == "id":
							e_id = value 
				for child in element.getchildren():
					ctag = str(child.tag).split('}')
					if ctag[1] == "title":
						e_title = child.text
					elif ctag[1] == "res":
						e_res = self.restoreString(child.text,0)
				Sonos_RT_FavStations.append ((e_res, e_title))
				self.plugin.debugLog(u"\tRadioTime Favorite Station: %s, %s, %s" % (e_id, e_title, e_res))
				list_count = list_count + 1
		indigo.server.log(u"Loaded RadioTime Favorite Stations... [%s]" % list_count)

	def getLineIn(self, dev, val):
		global Sonos_LineIn
		zoneName = dev.states['ZP_ZoneName']
		LocalUID = dev.states['ZP_LocalUID']
		for x in Sonos_LineIn:
			if x[0] == dev.id:
				Sonos_LineIn.remove([x[0], x[1]])
				indigo.server.log(u"LineIn: Removed: %s, %s" % (LocalUID, x[1]))
		Sonos_LineIn.append([dev.id, val + ":" + zoneName])
		indigo.server.log(u"LineIn: %s, %s:%s" % (LocalUID, val, zoneName))
		

	def getPandora(self, PandoraEmailAddress, PandoraPassword, PandoraNickname):
		global Sonos_Pandora
		list_count = 0
		#Sonos_Pandora = []
	
		pandora = Pandora()
		pandora.authenticate(PandoraEmailAddress, PandoraPassword)
		for station in pandora.get_station_list():
			Sonos_Pandora.append((station['stationId'], station['stationName'], PandoraEmailAddress, PandoraNickname))
			self.plugin.debugLog("\tPandora: %s, %s, %s" % (station['stationId'], station['stationName'], PandoraNickname))
			list_count = list_count + 1
		indigo.server.log(u"Loaded Pandora Stations for %s... [%s]" % (PandoraNickname, list_count))

	def getSiriusXM(self):
		global Sonos_SiriusXM
		#list_count = 0
		Sonos_SiriusXM = []

		# attempt 1 with website scraping
		#url = u"https://www.siriusxm.com/userservices/cl/en-us/xml/lineup/250/client/ump"
		#response = requests.get(url)
		#if response.ok:
		#	root = ET.fromstring(response.content)
		#	for Channel in root.findall('.//channels'):
		#		Sonos_SiriusXM.append((int(Channel.findtext('siriusChannelNo')), int(Channel.findtext('xmChannelNo')), Channel.findtext('channelKey'), Channel.findtext('contentId'), Channel.findtext('name')))
		#		self.plugin.debugLog('\tSiriusXM: %s|%s|%s|%s|%s' % (Channel.findtext('siriusChannelNo'), Channel.findtext('xmChannelNo'), Channel.findtext('channelKey'), Channel.findtext('contentId'), Channel.findtext('name')))
		#		list_count = list_count + 1
		#	Sonos_SiriusXM.sort(key=lambda x:x[0])
		#	indigo.server.log(u"Loaded SiriusXM Stations.. [%s]" % list_count)
		#else:
		#	self.plugin.errorLog("[%s] Error getting SiriusXM Channel Lineup" % time.asctime())

		# attempt 2 with MusicService
		#try:
		#	SiriusXMChannels = MusicService('SiriusXM').get_metadata(item='live_category:all_channels', count=500)
		#	for item in SiriusXMChannels:
		#		Sonos_SiriusXM.append((int(item.title.split('-')[0]), item.id, item.title))
		#		self.plugin.debugLog('\tSiriusXM: %s' % item.title)
		#	Sonos_SiriusXM.sort(key=lambda x:x[0])
		#	indigo.server.log(u"Loaded SiriusXM Stations.. [%s]" % len(SiriusXMChannels))
		#except:
		#	self.plugin.errorLog("[%s] Error getting SiriusXM Channel Lineup" % time.asctime())

		# attempt 3 with Sonos Music API
		headers = {'Content-Type': 'text/xml; charset="utf-8"',	 'Host': 'sonos.mountain.siriusxm.com', 'User-Agent': 'Indigo', 'CONNECTION': 'close', 'ACCEPT-ENCODING': 'gzip', 'ACCEPT-LANGUAGE': 'en-US'}
		base_url = 'http://sonos.mountain.siriusxm.com/server.php'
		namespaces = {'ns0': 'http://schemas.xmlsoap.org/soap/envelope', 'ns1': 'http://www.sonos.com/Services/1.1'}

		SoapMessage = (
			'<?xml version="1.0" encoding="utf-8"?>'
			'<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">'
				'<s:Header>'
					'<credentials xmlns="http://www.sonos.com/Services/1.1">'
						'<deviceId>' + self.SonosDeviceID + '</deviceId>'
			            '<deviceProvider>Sonos</deviceProvider>'
					'</credentials>'
				'</s:Header>'
			 	'<s:Body>'
			 		'<getSessionId xmlns="http://www.sonos.com/Services/1.1">'
						'<username>' + self.SiriusXMID + '</username>'
						'<password>' + self.SiriusXMPassword + '</password>'
					'</getSessionId>'
				'</s:Body>'
			'</s:Envelope>')

		headers['SOAPACTION'] = 'http://www.sonos.com/Services/1.1#getSessionId'
		headers['Content-Length'] = str(len(SoapMessage))

		try:
			response = requests.post(base_url, headers=headers, data=SoapMessage.encode("utf-8"))
			root = ET.fromstring(response.content)
			SessionID = root.findtext('.//ns1:getSessionIdResult', namespaces=namespaces)
		except Exception, e:
			print("[%s] SiriusXM SessionID error: %s" % (time.asctime(), e))

		SoapMessage = (
			'<?xml version="1.0" encoding="utf-8"?>'
			'<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">'
			    '<s:Header>'
			        '<credentials xmlns="http://www.sonos.com/Services/1.1">'
						'<sessionId>' + SessionID + '</sessionId>'
						'<deviceId>' + self.SonosDeviceID + '</deviceId>'
						'<deviceProvider>Sonos</deviceProvider>'
					'</credentials>'
				'</s:Header>'
				'<s:Body>'
					'<getMetadata xmlns="http://www.sonos.com/Services/1.1">'
						'<id>live_category:all_channels</id>'
						'<index>0</index>'
						'<count>500</count>'
					'</getMetadata>'
				'</s:Body>'
			'</s:Envelope>')

		headers['SOAPACTION'] = 'http://www.sonos.com/Services/1.1#getMetadata'
		headers['Content-Length'] = str(len(SoapMessage))

		try:
			response = requests.post(base_url, headers=headers, data=SoapMessage.encode("utf-8"))
			root = ET.fromstring(response.content)
			for item in root.findall('.//ns1:mediaMetadata', namespaces):
				xm_id = item.findtext('ns1:id', namespaces=namespaces)
				xm_title = item.findtext('ns1:title', namespaces=namespaces)
				Sonos_SiriusXM.append((int(xm_title.split('-')[0]), xm_id, xm_title))
				self.plugin.debugLog('\tSiriusXM: %s' % xm_title)
			Sonos_SiriusXM.sort(key=lambda x:x[0])
			indigo.server.log(u"Loaded SiriusXM Stations.. [%s]" % len(Sonos_SiriusXM))
		except:
			self.plugin.errorLog("[%s] Error getting SiriusXM Channel Lineup" % time.asctime())

	def getSoundFiles(self):
		global Sound_Files
		list_count = 0
		Sound_Files = []

		for f in listdir(self.SoundFilePath):
			if ".mp3" in f:
				Sound_Files.append(f)
				self.plugin.debugLog("\tSound File: %s" % f)
				list_count = list_count + 1

		indigo.server.log(u"Loaded Sound Files... [%s]" % list_count)
